import sys
import numpy as np
from phonopy.structure.symmetry import Symmetry
from phonopy.units import VaspToTHz
from phono3py.phonon3.triplets import (get_triplets_at_q,
                                       get_nosym_triplets_at_q,
                                       get_tetrahedra_vertices,
                                       get_triplets_integration_weights,
                                       occupation)
from phono3py.phonon3.interaction import set_phonon_c
from phono3py.phonon3.imag_self_energy import get_frequency_points
from phonopy.harmonic.dynamical_matrix import get_dynamical_matrix
from phonopy.structure.tetrahedron_method import TetrahedronMethod

class JointDos(object):
    def __init__(self,
                 mesh,
                 primitive,
                 supercell,
                 fc2,
                 nac_params=None,
                 nac_q_direction=None,
                 sigma=None,
                 cutoff_frequency=None,
                 frequency_step=None,
                 num_frequency_points=None,
                 temperatures=None,
                 frequency_factor_to_THz=VaspToTHz,
                 frequency_scale_factor=1.0,
                 is_mesh_symmetry=True,
                 symprec=1e-5,
                 filename=None,
                 log_level=False,
                 lapack_zheev_uplo='L'):

        self._grid_point = None
        self._mesh = np.array(mesh, dtype='intc')
        self._primitive = primitive
        self._supercell = supercell
        self._fc2 = fc2
        self._nac_params = nac_params
        self._nac_q_direction = None
        self.set_nac_q_direction(nac_q_direction)
        self._sigma = None
        self.set_sigma(sigma)

        if cutoff_frequency is None:
            self._cutoff_frequency = 0
        else:
            self._cutoff_frequency = cutoff_frequency
        self._frequency_step = frequency_step
        self._num_frequency_points = num_frequency_points
        self._temperatures = temperatures
        self._frequency_factor_to_THz = frequency_factor_to_THz
        self._frequency_scale_factor = frequency_scale_factor
        self._is_mesh_symmetry = is_mesh_symmetry
        self._symprec = symprec
        self._filename = filename
        self._log_level = log_level
        self._lapack_zheev_uplo = lapack_zheev_uplo

        self._num_band = self._primitive.get_number_of_atoms() * 3
        self._reciprocal_lattice = np.linalg.inv(self._primitive.get_cell())
        self._set_dynamical_matrix()
        self._symmetry = Symmetry(primitive, symprec)

        self._tetrahedron_method = None
        self._phonon_done = None
        self._frequencies = None
        self._eigenvectors = None
            
        self._joint_dos = None
        self._frequency_points = None

    def run(self):
        try:
            import phono3py._phono3py as phono3c
            self._run_c()
        except ImportError:
            print("Joint density of states in python is not implemented.")
            return None, None

    def get_joint_dos(self):
        return self._joint_dos

    def get_frequency_points(self):
        return self._frequency_points

    def get_phonons(self):
        return self._frequencies, self._eigenvectors, self._phonon_done

    def get_primitive(self):
        return self._primitive

    def get_mesh_numbers(self):
        return self._mesh
        
    def set_nac_q_direction(self, nac_q_direction=None):
        if nac_q_direction is not None:
            self._nac_q_direction = np.array(nac_q_direction, dtype='double')

    def set_sigma(self, sigma):
        if sigma is None:
            self._sigma = None
        else:
            self._sigma = float(sigma)

    def set_grid_point(self, grid_point):
        self._grid_point = grid_point
        self._set_triplets()
        num_grid = np.prod(len(self._grid_address))
        num_band = self._num_band
        if self._phonon_done is None:
            self._phonon_done = np.zeros(num_grid, dtype='byte')
            self._frequencies = np.zeros((num_grid, num_band), dtype='double')
            itemsize = self._frequencies.itemsize
            self._eigenvectors = np.zeros((num_grid, num_band, num_band),
                                          dtype=("c%d" % (itemsize * 2)))
            
        self._joint_dos = None
        self._frequency_points = None
        self.set_phonons(np.array([grid_point], dtype='intc'))

    def get_triplets_at_q(self):
        return self._triplets_at_q, self._weights_at_q
        
    def get_grid_address(self):
        return self._grid_address

    def get_bz_map(self):
        return self._bz_map
    
    def _run_c(self, lang='C'):
        if self._sigma is None:
            if lang == 'C':
                self._run_c_with_g()
            else:
                if self._temperatures is not None:
                    print("JDOS with phonon occupation numbers doesn't work "
                          "in this option.")
                self._run_py_tetrahedron_method()
        else:
            self._run_c_with_g()
                
    def _run_c_with_g(self):
        self.set_phonons(self._triplets_at_q.ravel())
        if self._sigma is None:
            f_max = np.max(self._frequencies) * 2
        else:
            f_max = np.max(self._frequencies) * 2 + self._sigma * 4
        f_max *= 1.005
        f_min = 0
        self._set_uniform_frequency_points(f_min, f_max)

        num_freq_points = len(self._frequency_points)
        num_mesh = np.prod(self._mesh)

        if self._temperatures is None:
            jdos = np.zeros((num_freq_points, 2), dtype='double')
        else:
            num_temps = len(self._temperatures)
            jdos = np.zeros((num_temps, num_freq_points, 2), dtype='double')
            occ_phonons = []
            for t in self._temperatures:
                freqs = self._frequencies[self._triplets_at_q[:, 1:]]
                occ_phonons.append(np.where(freqs > self._cutoff_frequency,
                                            occupation(freqs, t), 0))
        
        for i, freq_point in enumerate(self._frequency_points):
            g, _ = get_triplets_integration_weights(
                self,
                np.array([freq_point], dtype='double'),
                self._sigma,
                is_collision_matrix=True,
                neighboring_phonons=(i == 0))

            if self._temperatures is None:
                jdos[i, 1] = np.sum(
                    np.tensordot(g[0, :, 0], self._weights_at_q, axes=(0, 0)))
                gx = g[2] - g[0]
                jdos[i, 0] = np.sum(
                    np.tensordot(gx[:, 0], self._weights_at_q, axes=(0, 0)))
            else:
                for j, n in enumerate(occ_phonons):
                    for k, l in list(np.ndindex(g.shape[3:])):
                        jdos[j, i, 1] += np.dot(
                            (n[:, 0, k] + n[:, 1, l] + 1) *
                            g[0, :, 0, k, l], self._weights_at_q)
                        jdos[j, i, 0] += np.dot((n[:, 0, k] - n[:, 1, l]) *
                                                g[1, :, 0, k, l],
                                                self._weights_at_q)

        self._joint_dos = jdos / num_mesh
    
    def _run_py_tetrahedron_method(self):
        thm = TetrahedronMethod(self._reciprocal_lattice, mesh=self._mesh)
        self._vertices = get_tetrahedra_vertices(
            thm.get_tetrahedra(),
            self._mesh,
            self._triplets_at_q,
            self._grid_address,
            self._bz_map)
        self.set_phonons(self._vertices.ravel())
        f_max = np.max(self._frequencies) * 2
        f_max *= 1.005
        f_min = 0
        self._set_uniform_frequency_points(f_min, f_max)

        num_freq_points = len(self._frequency_points)
        jdos = np.zeros((num_freq_points, 2), dtype='double')
        for vertices, w in zip(self._vertices, self._weights_at_q):
            for i, j in list(np.ndindex(self._num_band, self._num_band)):
                f1 = self._frequencies[vertices[0], i]
                f2 = self._frequencies[vertices[1], j]
                thm.set_tetrahedra_omegas(f1 + f2)
                thm.run(self._frequency_points)
                iw = thm.get_integration_weight()
                jdos[:, 1] += iw * w

                thm.set_tetrahedra_omegas(f1 - f2)
                thm.run(self._frequency_points)
                iw = thm.get_integration_weight()
                jdos[:, 0] += iw * w

                thm.set_tetrahedra_omegas(-f1 + f2)
                thm.run(self._frequency_points)
                iw = thm.get_integration_weight()
                jdos[:, 0] += iw * w

        self._joint_dos = jdos / np.prod(self._mesh)

    def _set_dynamical_matrix(self):
        self._dm = get_dynamical_matrix(
            self._fc2,
            self._supercell,
            self._primitive,
            nac_params=self._nac_params,
            frequency_scale_factor=self._frequency_scale_factor,
            symprec=self._symprec)
        
    def _set_triplets(self):
        if not self._is_mesh_symmetry:
            if self._log_level:
                print("Triplets at q without considering symmetry")
                sys.stdout.flush()
            
            (self._triplets_at_q,
             self._weights_at_q,
             self._grid_address,
             self._bz_map,
             map_triplets,
             map_q) = get_nosym_triplets_at_q(
                 self._grid_point,
                 self._mesh,
                 self._reciprocal_lattice,
                 with_bz_map=True)
        else:
            (self._triplets_at_q,
             self._weights_at_q,
             self._grid_address,
             self._bz_map,
             map_triplets,
             map_q) = get_triplets_at_q(
                 self._grid_point,
                 self._mesh,
                 self._symmetry.get_pointgroup_operations(),
                 self._reciprocal_lattice)

    def set_phonons(self, grid_points):
        set_phonon_c(self._dm,
                     self._frequencies,
                     self._eigenvectors,
                     self._phonon_done,
                     grid_points,
                     self._grid_address,
                     self._mesh,
                     self._frequency_factor_to_THz,
                     self._nac_q_direction,
                     self._lapack_zheev_uplo)

    def set_frequency_points(self, frequency_points):
        self._frequency_points = np.array(frequency_points, dtype='double')

    def _set_uniform_frequency_points(self, f_min, f_max):
        if self._frequency_points is None:
            self._frequency_points = get_frequency_points(
                f_min,
                f_max,
                frequency_step=self._frequency_step,
                num_frequency_points=self._num_frequency_points)



import numpy as np
from phonopy.harmonic.dynamical_matrix import get_dynamical_matrix
from phonopy.units import (VaspToTHz, Hbar, EV, Angstrom, THz, AMU,
                           PlanckConstant)
from phono3py.phonon.solver import set_phonon_c, set_phonon_py
from phono3py.phonon3.real_to_reciprocal import RealToReciprocal
from phono3py.phonon3.reciprocal_to_normal import ReciprocalToNormal
from phono3py.phonon3.triplets import (get_triplets_at_q,
                                       get_nosym_triplets_at_q,
                                       get_bz_grid_address)

class Interaction(object):
    def __init__(self,
                 supercell,
                 primitive,
                 mesh,
                 symmetry,
                 fc3=None,
                 band_indices=None,
                 constant_averaged_interaction=None,
                 frequency_factor_to_THz=VaspToTHz,
                 unit_conversion=None,
                 is_mesh_symmetry=True,
                 symmetrize_fc3_q=False,
                 cutoff_frequency=None,
                 lapack_zheev_uplo='L'):
        self._fc3 = fc3
        self._supercell = supercell
        self._primitive = primitive
        self._mesh = np.array(mesh, dtype='intc')
        self._symmetry = symmetry

        self._band_indices = None
        self._set_band_indices(band_indices)
        self._constant_averaged_interaction = constant_averaged_interaction
        self._frequency_factor_to_THz = frequency_factor_to_THz

        # Unit to eV^2
        if unit_conversion is None:
            num_grid = np.prod(self._mesh)
            self._unit_conversion = ((Hbar * EV) ** 3 / 36 / 8
                                     * EV ** 2 / Angstrom ** 6
                                     / (2 * np.pi * THz) ** 3
                                     / AMU ** 3 / num_grid
                                     / EV ** 2)
        else:
            self._unit_conversion = unit_conversion
        if cutoff_frequency is None:
            self._cutoff_frequency = 0
        else:
            self._cutoff_frequency = cutoff_frequency
        self._is_mesh_symmetry = is_mesh_symmetry
        self._symmetrize_fc3_q = symmetrize_fc3_q
        self._lapack_zheev_uplo = lapack_zheev_uplo

        self._symprec = symmetry.get_symmetry_tolerance()

        self._grid_point = None
        self._triplets_at_q = None
        self._weights_at_q = None
        self._triplets_map_at_q = None
        self._ir_map_at_q = None
        self._grid_address = None
        self._bz_map = None
        self._interaction_strength = None

        self._phonon_done = None
        self._frequencies = None
        self._eigenvectors = None
        self._dm = None
        self._nac_q_direction = None

        self._band_index_count = 0

        svecs, multiplicity = self._primitive.get_smallest_vectors()
        self._smallest_vectors = svecs
        self._multiplicity = multiplicity
        self._masses = np.array(self._primitive.get_masses(), dtype='double')
        self._p2s = self._primitive.get_primitive_to_supercell_map()
        self._s2p = self._primitive.get_supercell_to_primitive_map()

        self._allocate_phonon()

    def run(self, lang='C', g_zero=None):
        num_band = self._primitive.get_number_of_atoms() * 3
        num_triplets = len(self._triplets_at_q)

        self._interaction_strength = np.empty(
            (num_triplets, len(self._band_indices), num_band, num_band),
            dtype='double')
        if self._constant_averaged_interaction is None:
            self._interaction_strength[:] = 0
            if lang == 'C':
                self._run_c(g_zero)
            else:
                self._run_py()
        else:
            num_grid = np.prod(self._mesh)
            self._interaction_strength[:] = (
                self._constant_averaged_interaction / num_grid)

    def get_interaction_strength(self):
        return self._interaction_strength

    def get_mesh_numbers(self):
        return self._mesh

    def get_phonons(self):
        return self._frequencies, self._eigenvectors, self._phonon_done

    def get_fc3(self):
        return self._fc3

    def get_dynamical_matrix(self):
        return self._dm

    def get_primitive(self):
        return self._primitive

    def get_supercell(self):
        return self._supercell

    def get_triplets_at_q(self):
        return (self._triplets_at_q,
                self._weights_at_q,
                self._triplets_map_at_q,
                self._ir_map_at_q)

    def get_grid_address(self):
        return self._grid_address

    def get_bz_map(self):
        return self._bz_map

    def get_band_indices(self):
        return self._band_indices

    def get_frequency_factor_to_THz(self):
        return self._frequency_factor_to_THz

    def get_lapack_zheev_uplo(self):
        return self._lapack_zheev_uplo

    def get_cutoff_frequency(self):
        return self._cutoff_frequency

    def get_averaged_interaction(self):
        v = self._interaction_strength
        w = self._weights_at_q
        v_sum = v.sum(axis=2).sum(axis=2)
        num_band = self._primitive.get_number_of_atoms() * 3
        return np.dot(w, v_sum) / num_band ** 2

    def get_primitive_and_supercell_correspondence(self):
        return (self._smallest_vectors,
                self._multiplicity,
                self._p2s,
                self._s2p,
                self._masses)

    def get_nac_q_direction(self):
        return self._nac_q_direction

    def get_unit_conversion_factor(self):
        return self._unit_conversion

    def get_constant_averaged_interaction(self):
        return self._constant_averaged_interaction

    def set_grid_point(self, grid_point, stores_triplets_map=False):
        reciprocal_lattice = np.linalg.inv(self._primitive.get_cell())
        if not self._is_mesh_symmetry:
            (triplets_at_q,
             weights_at_q,
             grid_address,
             bz_map,
             triplets_map_at_q,
             ir_map_at_q) = get_nosym_triplets_at_q(
                 grid_point,
                 self._mesh,
                 reciprocal_lattice,
                 stores_triplets_map=stores_triplets_map)
        else:
            (triplets_at_q,
             weights_at_q,
             grid_address,
             bz_map,
             triplets_map_at_q,
             ir_map_at_q)= get_triplets_at_q(
                 grid_point,
                 self._mesh,
                 self._symmetry.get_pointgroup_operations(),
                 reciprocal_lattice,
                 stores_triplets_map=stores_triplets_map)

        # Special treatment of symmetry is applied when q_direction is used.
        if (grid_address[grid_point] == 0).all():
            self._phonon_done[grid_point] = 0
            self.set_phonons(np.array([0], dtype='intc'))
            if self._nac_q_direction is not None:
                rotations = []
                for r in self._symmetry.get_pointgroup_operations():
                    dq = self._nac_q_direction
                    dq /= np.linalg.norm(dq)
                    diff = np.dot(dq, r) - dq
                    if (abs(diff) < 1e-5).all():
                        rotations.append(r)
                (triplets_at_q,
                 weights_at_q,
                 grid_address,
                 bz_map,
                 triplets_map_at_q,
                 ir_map_at_q) = get_triplets_at_q(
                     grid_point,
                     self._mesh,
                     np.array(rotations, dtype='intc', order='C'),
                     reciprocal_lattice,
                     is_time_reversal=False,
                     stores_triplets_map=stores_triplets_map)

        for triplet in triplets_at_q:
            sum_q = (grid_address[triplet]).sum(axis=0)
            if (sum_q % self._mesh != 0).any():
                print("============= Warning ==================")
                print("%s" % triplet)
                for tp in triplet:
                    print("%s %s" %
                          (grid_address[tp],
                           np.linalg.norm(
                               np.dot(reciprocal_lattice,
                                      grid_address[tp] /
                                      self._mesh.astype('double')))))
                print("%s" % sum_q)
                print("============= Warning ==================")

        self._grid_point = grid_point
        self._triplets_at_q = triplets_at_q
        self._weights_at_q = weights_at_q
        self._triplets_map_at_q = triplets_map_at_q
        # self._grid_address = grid_address
        # self._bz_map = bz_map
        self._ir_map_at_q = ir_map_at_q

        # set_phonons is unnecessary now because all phonons are calculated in
        # set_dynamical_matrix, though Gamma-point is an exception.
        # self.set_phonons(self._triplets_at_q.ravel())

    def set_dynamical_matrix(self,
                             fc2,
                             supercell,
                             primitive,
                             nac_params=None,
                             frequency_scale_factor=None,
                             decimals=None):
        self._dm = get_dynamical_matrix(
            fc2,
            supercell,
            primitive,
            nac_params=nac_params,
            frequency_scale_factor=frequency_scale_factor,
            decimals=decimals,
            symprec=self._symprec)
        self.set_phonons(np.arange(len(self._grid_address), dtype='intc'))
        if (self._grid_address[0] == 0).all():
            if np.sum(self._frequencies[0] < self._cutoff_frequency) < 3:
                for i, f in enumerate(self._frequencies[0, :3]):
                    if not (f < self._cutoff_frequency):
                        self._frequencies[0, i] = 0
                        print("=" * 26 + " Warning " + "=" * 26)
                        print(" Phonon frequency of band index %d at Gamma "
                              "is calculated to be %f." % (i + 1, f))
                        print(" But this frequency is forced to be zero.")
                        print("=" * 61)

    def set_nac_q_direction(self, nac_q_direction=None):
        if nac_q_direction is not None:
            self._nac_q_direction = np.array(nac_q_direction, dtype='double')

    def set_phonon_data(self, frequencies, eigenvectors, grid_address):
        if grid_address.shape != self._grid_address.shape:
            print("=" * 26 + " Warning " + "=" * 26)
            print("Input grid address size is inconsistent. "
                  "Setting phonons faild.")
            print("=" * 26 + " Warning " + "=" * 26)
            return False

        if (self._grid_address - grid_address).all():
            print("=" * 26 + " Warning " + "=" * 26)
            print("Input grid addresses are inconsistent. "
                  "Setting phonons faild.")
            print("=" * 26 + " Warning " + "=" * 26)
            return False
        else:
            self._phonon_done[:] = 1
            self._frequencies[:] = frequencies
            self._eigenvectors[:] = eigenvectors
            return True

    def set_phonons(self, grid_points):
        # for i, grid_triplet in enumerate(self._triplets_at_q):
        #     for gp in grid_triplet:
        #         self._set_phonon_py(gp)
        self._set_phonon_c(grid_points)

    def delete_interaction_strength(self):
        self._interaction_strength = None

    def _set_band_indices(self, band_indices):
        num_band = self._primitive.get_number_of_atoms() * 3
        if band_indices is None:
            self._band_indices = np.arange(num_band, dtype='intc')
        else:
            self._band_indices = np.array(band_indices, dtype='intc')

    def _run_c(self, g_zero):
        import phono3py._phono3py as phono3c

        num_band = self._primitive.get_number_of_atoms() * 3

        if g_zero is None or self._symmetrize_fc3_q:
            _g_zero = np.zeros(self._interaction_strength.shape,
                               dtype='byte', order='C')
        else:
            _g_zero = g_zero

        phono3c.interaction(self._interaction_strength,
                            _g_zero,
                            self._frequencies,
                            self._eigenvectors,
                            self._triplets_at_q,
                            self._grid_address,
                            self._mesh,
                            self._fc3,
                            self._smallest_vectors,
                            self._multiplicity,
                            self._masses,
                            self._p2s,
                            self._s2p,
                            self._band_indices,
                            self._symmetrize_fc3_q,
                            self._cutoff_frequency)
        self._interaction_strength *= self._unit_conversion

    def _set_phonon_c(self, grid_points):
        set_phonon_c(self._dm,
                     self._frequencies,
                     self._eigenvectors,
                     self._phonon_done,
                     grid_points,
                     self._grid_address,
                     self._mesh,
                     self._frequency_factor_to_THz,
                     self._nac_q_direction,
                     self._lapack_zheev_uplo)

    def _run_py(self):
        r2r = RealToReciprocal(self._fc3,
                               self._supercell,
                               self._primitive,
                               self._mesh,
                               symprec=self._symprec)
        r2n = ReciprocalToNormal(self._primitive,
                                 self._frequencies,
                                 self._eigenvectors,
                                 self._band_indices,
                                 cutoff_frequency=self._cutoff_frequency)

        for i, grid_triplet in enumerate(self._triplets_at_q):
            print("%d / %d" % (i + 1, len(self._triplets_at_q)))
            r2r.run(self._grid_address[grid_triplet])
            fc3_reciprocal = r2r.get_fc3_reciprocal()
            for gp in grid_triplet:
                self._set_phonon_py(gp)
            r2n.run(fc3_reciprocal, grid_triplet)
            self._interaction_strength[i] = np.abs(
                r2n.get_reciprocal_to_normal()) ** 2 * self._unit_conversion

    def _set_phonon_py(self, grid_point):
        set_phonon_py(grid_point,
                      self._phonon_done,
                      self._frequencies,
                      self._eigenvectors,
                      self._grid_address,
                      self._mesh,
                      self._dm,
                      self._frequency_factor_to_THz,
                      self._lapack_zheev_uplo)

    def _allocate_phonon(self):
        primitive_lattice = np.linalg.inv(self._primitive.get_cell())
        self._grid_address, self._bz_map = get_bz_grid_address(
            self._mesh, primitive_lattice, with_boundary=True)
        num_band = self._primitive.get_number_of_atoms() * 3
        num_grid = len(self._grid_address)
        self._phonon_done = np.zeros(num_grid, dtype='byte')
        self._frequencies = np.zeros((num_grid, num_band), dtype='double')
        itemsize = self._frequencies.itemsize
        self._eigenvectors = np.zeros((num_grid, num_band, num_band),
                                      dtype=("c%d" % (itemsize * 2)))



import numpy as np
from phonopy.harmonic.displacement import (get_least_displacements,
                                           directions_axis, get_displacement,
                                           is_minus_displacement)
from phonopy.structure.cells import get_equivalent_smallest_vectors

def direction_to_displacement(dataset,
                              distance,
                              supercell,
                              cutoff_distance=None):
    lattice = supercell.get_cell().T
    new_dataset = {}
    new_dataset['natom'] = supercell.get_number_of_atoms()
    if cutoff_distance is not None:
        new_dataset['cutoff_distance'] = cutoff_distance
    new_first_atoms = []
    for first_atoms in dataset:
        atom1 = first_atoms['number']
        direction1 = first_atoms['direction']
        disp_cart1 = np.dot(direction1, lattice.T)
        disp_cart1 *= distance / np.linalg.norm(disp_cart1)
        new_second_atoms = []
        for second_atom in first_atoms['second_atoms']:
            atom2 = second_atom['number']
            pair_distance = second_atom['distance']
            included = (cutoff_distance is None or
                        pair_distance < cutoff_distance)
            for direction2 in second_atom['directions']:
                disp_cart2 = np.dot(direction2, lattice.T)
                disp_cart2 *= distance / np.linalg.norm(disp_cart2)
                if cutoff_distance is None:
                    new_second_atoms.append({'number': atom2,
                                             'direction': direction2,
                                             'displacement': disp_cart2,
                                             'pair_distance': pair_distance})
                else:
                    new_second_atoms.append({'number': atom2,
                                             'direction': direction2,
                                             'displacement': disp_cart2,
                                             'pair_distance': pair_distance,
                                             'included': included})
        new_first_atoms.append({'number': atom1,
                                'direction': direction1,
                                'displacement': disp_cart1,
                                'second_atoms': new_second_atoms})
    new_dataset['first_atoms'] = new_first_atoms
    
    return new_dataset

def get_third_order_displacements(cell,
                                  symmetry,
                                  is_plusminus='auto',
                                  is_diagonal=False):
    # Atoms 1, 2, and 3 are defined as follows:
    #
    # Atom 1: The first displaced atom. Third order force constant
    #         between Atoms 1, 2, and 3 is calculated.
    # Atom 2: The second displaced atom. Second order force constant
    #         between Atoms 2 and 3 is calculated.
    # Atom 3: Force is mesuared on this atom.

    positions = cell.get_scaled_positions()
    lattice = cell.get_cell().T

    # Least displacements for third order force constants in yaml file
    #
    # Data structure
    # [{'number': atom1,
    #   'displacement': [0.00000, 0.007071, 0.007071],
    #   'second_atoms': [ {'number': atom2,
    #                      'displacements': [[0.007071, 0.000000, 0.007071],
    #                                        [-0.007071, 0.000000, -0.007071]
    #                                        ,...]},
    #                     {'number': ... } ] },
    #  {'number': atom1, ... } ]

    # Least displacements of first atoms (Atom 1) are searched by
    # using respective site symmetries of the original crystal.
    disps_first = get_least_displacements(symmetry,
                                          is_plusminus=is_plusminus,
                                          is_diagonal=False)

    symprec = symmetry.get_symmetry_tolerance()

    dds = []
    for disp in disps_first:
        atom1 = disp[0]
        disp1 = disp[1:4]
        site_sym = symmetry.get_site_symmetry(atom1)

        dds_atom1 = {'number': atom1,
                     'direction': disp1,
                     'second_atoms': []}

        # Reduced site symmetry at the first atom with respect to 
        # the displacement of the first atoms.
        reduced_site_sym = get_reduced_site_symmetry(site_sym, disp1, symprec)
        # Searching orbits (second atoms) with respect to
        # the first atom and its reduced site symmetry.
        second_atoms = get_least_orbits(atom1,
                                        cell,
                                        reduced_site_sym,
                                        symprec)

        for atom2 in second_atoms:
            dds_atom2 = get_next_displacements(atom1,
                                               atom2,
                                               reduced_site_sym,
                                               lattice,
                                               positions,
                                               symprec,
                                               is_diagonal)

            min_distance = np.linalg.norm(
                np.dot(lattice, get_equivalent_smallest_vectors(
                    atom1,
                    atom2,
                    cell,
                    lattice.T,
                    symprec)[0]))
            dds_atom2['distance'] = min_distance
            dds_atom1['second_atoms'].append(dds_atom2)
        dds.append(dds_atom1)
    
    return dds

def get_next_displacements(atom1,
                           atom2,
                           reduced_site_sym,
                           lattice,
                           positions,
                           symprec,
                           is_diagonal):
    # Bond symmetry between first and second atoms.
    reduced_bond_sym = get_bond_symmetry(
        reduced_site_sym,
        lattice,
        positions, 
        atom1,
        atom2,
        symprec)

    # Since displacement of first atom breaks translation
    # symmetry, the crystal symmetry is reduced to point
    # symmetry and it is equivalent to the site symmetry
    # on the first atom. Therefore site symmetry on the 
    # second atom with the displacement is equivalent to
    # this bond symmetry.
    if is_diagonal:
        disps_second = get_displacement(reduced_bond_sym)
    else:
        disps_second = get_displacement(reduced_bond_sym, directions_axis)
    dds_atom2 = {'number': atom2, 'directions': []}
    for disp2 in disps_second:
        dds_atom2['directions'].append(disp2)
        if is_minus_displacement(disp2, reduced_bond_sym):
            dds_atom2['directions'].append(-disp2)

    return dds_atom2


def get_reduced_site_symmetry(site_sym, direction, symprec=1e-5):
    reduced_site_sym = []
    for rot in site_sym:
        if (abs(direction - np.dot(direction, rot.T)) < symprec).all():
            reduced_site_sym.append(rot)
    return np.array(reduced_site_sym, dtype='intc')

def get_bond_symmetry(site_symmetry,
                      lattice,
                      positions,
                      atom_center,
                      atom_disp,
                      symprec=1e-5):
    """
    Bond symmetry is the symmetry operations that keep the symmetry
    of the cell containing two fixed atoms.
    """
    bond_sym = []
    pos = positions
    for rot in site_symmetry:
        rot_pos = (np.dot(pos[atom_disp] - pos[atom_center], rot.T) +
                   pos[atom_center])
        diff = pos[atom_disp] - rot_pos
        diff -= np.rint(diff)
        dist = np.linalg.norm(np.dot(lattice, diff))
        if dist < symprec:
            bond_sym.append(rot)

    return np.array(bond_sym)

def get_least_orbits(atom_index, cell, site_symmetry, symprec=1e-5):
    """Find least orbits for a centering atom"""
    orbits = _get_orbits(atom_index, cell, site_symmetry, symprec)
    mapping = np.arange(cell.get_number_of_atoms())

    for i, orb in enumerate(orbits):
        for num in np.unique(orb):
            if mapping[num] > mapping[i]:
                mapping[num] = mapping[i]

    return np.unique(mapping)

def _get_orbits(atom_index, cell, site_symmetry, symprec=1e-5):
    lattice = cell.get_cell().T
    positions = cell.get_scaled_positions()
    center = positions[atom_index]

    # orbits[num_atoms, num_site_sym]
    orbits = []
    for pos in positions:
        mapping = []

        for rot in site_symmetry:
            rot_pos = np.dot(pos - center, rot.T) + center

            for i, pos in enumerate(positions):
                diff = pos - rot_pos
                diff -= np.rint(diff)
                dist = np.linalg.norm(np.dot(lattice, diff))
                if dist < symprec:
                    mapping.append(i)
                    break

        if len(mapping) < len(site_symmetry):
            print("Site symmetry is broken.")
            raise ValueError
        else:
            orbits.append(mapping)
        
    return np.array(orbits)




import sys
import numpy as np
from phonopy.harmonic.dynamical_matrix import (DynamicalMatrix,
                                               DynamicalMatrixNAC)
from phonopy.structure.cells import get_supercell
from phono3py.file_IO import write_fc3_dat, write_fc2_dat
from phonopy.units import VaspToTHz
from phonopy.structure.grid_points import get_qpoints

def get_gruneisen_parameters(fc2,
                             fc3,
                             supercell,
                             primitive,
                             band_paths,
                             mesh,
                             qpoints,
                             nac_params=None,
                             nac_q_direction=None,
                             ion_clamped=False,
                             factor=None,
                             symprec=1e-5,
                             output_filename=None,
                             log_level=True):
    if log_level:
        print("-" * 23 + " Phonon Gruneisen parameter " + "-" * 23)
        if mesh is not None:
            print("Mesh sampling: [ %d %d %d ]" % tuple(mesh))
        elif band_paths is not None:
            print("Paths in reciprocal reduced coordinates:")
            for path in band_paths:
                print("[%5.2f %5.2f %5.2f] --> [%5.2f %5.2f %5.2f]" %
                      (tuple(path[0]) + tuple(path[-1])))
        if ion_clamped:
            print("To be calculated with ion clamped.")

        sys.stdout.flush()

    gruneisen = Gruneisen(fc2,
                          fc3,
                          supercell,
                          primitive,
                          nac_params=nac_params,
                          nac_q_direction=nac_q_direction,
                          ion_clamped=ion_clamped,
                          factor=factor,
                          symprec=symprec)

    if mesh is not None:
        gruneisen.set_sampling_mesh(mesh, is_gamma_center=True)
    elif band_paths is not None:
        gruneisen.set_band_structure(band_paths)
    elif qpoints is not None:
        gruneisen.set_qpoints(qpoints)
    gruneisen.run()

    if output_filename is None:
        filename = 'gruneisen'
    else:
        filename = 'gruneisen.' + output_filename
    gruneisen.write(filename=filename)

class Gruneisen(object):
    def __init__(self,
                 fc2,
                 fc3,
                 supercell,
                 primitive,
                 nac_params=None,
                 nac_q_direction=None,
                 ion_clamped=False,
                 factor=VaspToTHz,
                 symprec=1e-5):
        self._fc2 = fc2
        self._fc3 = fc3
        self._scell = supercell
        self._pcell = primitive
        self._ion_clamped = ion_clamped
        self._factor = factor
        self._symprec = symprec
        if nac_params is None:
            self._dm = DynamicalMatrix(self._scell,
                                       self._pcell,
                                       self._fc2,
                                       symprec=self._symprec)
        else:
            self._dm = DynamicalMatrixNAC(self._scell,
                                          self._pcell,
                                          self._fc2,
                                          symprec=self._symprec)
            self._dm.set_nac_params(nac_params)
        self._nac_q_direction = nac_q_direction

        (self._shortest_vectors,
         self._multiplicity) = self._pcell.get_smallest_vectors()

        if self._ion_clamped:
            num_atom_prim = self._pcell.get_number_of_atoms()
            self._X = np.zeros((num_atom_prim, 3, 3, 3), dtype=float)
        else:
            self._X = self._get_X()
        self._dPhidu = self._get_dPhidu()

        self._gruneisen_parameters = None
        self._frequencies = None
        self._qpoints = None
        self._mesh = None
        self._band_paths = None
        self._band_distances = None
        self._run_mode = None
        self._weights = None

    def run(self):
        if self._run_mode == 'band':
            (self._gruneisen_parameters,
             self._frequencies) = self._calculate_band_paths()
        elif self._run_mode == 'qpoints' or self._run_mode == 'mesh':
            (self._gruneisen_parameters,
             self._frequencies) = self._calculate_at_qpoints(self._qpoints)
        else:
            sys.stderr.write('Q-points are not specified.\n')

    def get_gruneisen_parameters(self):
        return self._gruneisen_parameters

    def set_qpoints(self, qpoints):
        self._run_mode = 'qpoints'
        self._qpoints = qpoints

    def set_sampling_mesh(self,
                          mesh,
                          shift=None,
                          is_gamma_center=False):
        self._run_mode = 'mesh'
        self._mesh = mesh
        self._qpoints, self._weights = get_qpoints(
            self._mesh,
            np.linalg.inv(self._pcell.get_cell()),
            q_mesh_shift=shift,
            is_gamma_center=is_gamma_center)

    def set_band_structure(self, paths):
        self._run_mode = 'band'
        self._band_paths = paths
        rec_lattice = np.linalg.inv(self._pcell.get_cell())
        self._band_distances = []
        for path in paths:
            distances_at_path = [0.]
            for i in range(len(path) - 1):
                distances_at_path.append(np.linalg.norm(
                        np.dot(rec_lattice, path[i + 1] - path[i])) +
                                         distances_at_path[-1])
            self._band_distances.append(distances_at_path)

    def write(self, filename="gruneisen"):
        if self._gruneisen_parameters is not None:
            if self._run_mode == 'band':
                self._write_band_yaml(filename + ".yaml")
            elif self._run_mode == 'qpoints' or self._run_mode == 'mesh':
                # self._write_mesh_yaml(filename + ".yaml")
                self._write_mesh_hdf5(filename + ".hdf5")

    def _write_mesh_yaml(self, filename):
        with open(filename, 'w') as f:
            if self._run_mode == 'mesh':
                f.write("mesh: [ %5d, %5d, %5d ]\n" % tuple(self._mesh))
            f.write("nqpoint: %d\n" % len(self._qpoints))
            f.write("phonon:\n")
            for i, (q, g_at_q, freqs_at_q) in enumerate(
                zip(self._qpoints,
                    self._gruneisen_parameters,
                    self._frequencies)):
                f.write("- q-position: [ %10.7f, %10.7f, %10.7f ]\n" % tuple(q))
                if self._weights is not None:
                    f.write("  multiplicity: %d\n" % self._weights[i])
                f.write("  band:\n")
                for j, (g, freq) in enumerate(zip(g_at_q, freqs_at_q)):
                    f.write("  - # %d\n" % (j + 1))
                    f.write("    frequency: %15.10f\n" % freq)
                    f.write("    gruneisen: %15.10f\n" % (g.trace() / 3))
                    f.write("    gruneisen_tensor:\n")
                    for g_xyz in g:
                        f.write("    - [ %10.7f, %10.7f, %10.7f ]\n" %
                                tuple(g_xyz))

    def _write_band_yaml(self, filename):
        with open(filename, 'w') as f:
            f.write("path:\n\n")
            for path, distances, gs, fs in zip(self._band_paths,
                                               self._band_distances,
                                               self._gruneisen_parameters,
                                               self._frequencies):
                f.write("- nqpoint: %d\n" % len(path))
                f.write("  phonon:\n")
                for i, (q, d, g_at_q, freqs_at_q) in enumerate(
                    zip(path, distances, gs, fs)):
                    f.write("  - q-position: [ %10.7f, %10.7f, %10.7f ]\n"
                            % tuple(q))
                    f.write("    distance: %10.7f\n" % d)
                    f.write("    band:\n")
                    for j, (g, freq) in enumerate(zip(g_at_q, freqs_at_q)):
                        f.write("    - # %d\n" % (j + 1))
                        f.write("      frequency: %15.10f\n" % freq)
                        f.write("      gruneisen: %15.10f\n" % (g.trace() / 3))
                        f.write("      gruneisen_tensor:\n")
                        for g_xyz in g:
                            f.write("      - [ %10.7f, %10.7f, %10.7f ]\n" %
                                    tuple(g_xyz))
                    f.write("\n")

    def _write_mesh_hdf5(self, filename="gruneisen.hdf5"):
        import h5py

        g = self._gruneisen_parameters
        gruneisen = np.array(
            (g[:, :, 0, 0] + g[:, :, 1, 1] + g[:, :, 2, 2]) / 3,
            dtype='double', order='C')

        with h5py.File(filename, 'w') as w:
            w.create_dataset('mesh', data=self._mesh)
            w.create_dataset('gruneisen', data=gruneisen)
            w.create_dataset('gruneisen_tensor',
                             data=self._gruneisen_parameters)
            w.create_dataset('weight', data=self._weights)
            w.create_dataset('frequency', data=self._frequencies)
            w.create_dataset('qpoint', data=self._qpoints)

    def _calculate_at_qpoints(self, qpoints):
        gruneisen_parameters = []
        frequencies = []
        for i, q in enumerate(qpoints):
            if self._dm.is_nac():
                if (np.abs(q) < 1e-5).all(): # If q is almost at Gamma
                    if self._run_mode == 'band':
                        # Direction estimated from neighboring point
                        if i > 0:
                            q_dir = qpoints[i] - qpoints[i - 1]
                        elif i == 0 and len(qpoints) > 1:
                            q_dir = qpoints[i + 1] - qpoints[i]
                        else:
                            q_dir = None
                        g, omega2 = self._get_gruneisen_tensor(
                            q, nac_q_direction=q_dir)
                    else: # Specified q-vector
                        g, omega2 = self._get_gruneisen_tensor(
                            q, nac_q_direction=self._nac_q_direction)
                else: # If q is away from Gamma-point, then q-vector
                    g, omega2 = self._get_gruneisen_tensor(q, nac_q_direction=q)
            else: # Without NAC
                g, omega2 = self._get_gruneisen_tensor(q)
            gruneisen_parameters.append(g)
            frequencies.append(
                np.sqrt(abs(omega2)) * np.sign(omega2) * self._factor)

        return (np.array(gruneisen_parameters, dtype='double', order='C'),
                np.array(frequencies, dtype='double', order='C'))

    def _calculate_band_paths(self):
        gruneisen_parameters = []
        frequencies = []
        for path in self._band_paths:
            (gruneisen_at_path,
             frequencies_at_path) = self._calculate_at_qpoints(path)
            gruneisen_parameters.append(gruneisen_at_path)
            frequencies.append(frequencies_at_path)

        return gruneisen_parameters, frequencies

    def _get_gruneisen_tensor(self, q, nac_q_direction=None):
        if nac_q_direction is None:
            self._dm.set_dynamical_matrix(q)
        else:
            self._dm.set_dynamical_matrix(q, nac_q_direction)
        omega2, w = np.linalg.eigh(self._dm.get_dynamical_matrix())
        g = np.zeros((len(omega2), 3, 3), dtype=float)
        num_atom_prim = self._pcell.get_number_of_atoms()
        dDdu = self._get_dDdu(q)

        for s in range(len(omega2)):
            if (np.abs(q) < 1e-5).all() and s < 3:
                continue
            for i in range(3):
                for j in range(3):
                    for nu in range(num_atom_prim):
                        for pi in range(num_atom_prim):
                            g[s] += (w[nu * 3 + i, s].conjugate() *
                                     dDdu[nu, pi, i, j] * w[pi * 3 + j, s]).real

            g[s] *= -1.0 / 2 / omega2[s]

        return g, omega2

    def _get_dDdu(self, q):
        num_atom_prim = self._pcell.get_number_of_atoms()
        num_atom_super = self._scell.get_number_of_atoms()
        p2s = self._pcell.get_primitive_to_supercell_map()
        s2p = self._pcell.get_supercell_to_primitive_map()
        vecs = self._shortest_vectors
        multi = self._multiplicity
        m = self._pcell.get_masses()
        dPhidu = self._dPhidu
        dDdu = np.zeros((num_atom_prim, num_atom_prim, 3, 3, 3, 3), dtype=complex)

        for nu in range(num_atom_prim):
            for pi, p in enumerate(p2s):
                for Ppi, s in enumerate(s2p):
                    if not s==p:
                        continue
                    phase = np.exp(2j * np.pi * np.dot(
                            vecs[Ppi,nu,:multi[Ppi, nu], :], q)
                                   ).sum() / multi[Ppi, nu]
                    dDdu[nu, pi] += phase * dPhidu[nu, Ppi]
                dDdu[nu, pi] /= np.sqrt(m[nu] * m[pi])

        return dDdu

    def _get_dPhidu(self):
        fc3 = self._fc3
        num_atom_prim = self._pcell.get_number_of_atoms()
        num_atom_super = self._scell.get_number_of_atoms()
        p2s = self._pcell.get_primitive_to_supercell_map()
        dPhidu = np.zeros((num_atom_prim, num_atom_super, 3, 3, 3, 3),
                          dtype=float)

        for nu in range(num_atom_prim):
            Y = self._get_Y(nu)
            for pi in range(num_atom_super):
                for i in range(3):
                    for j in range(3):
                        for k in range(3):
                            for l in range(3):
                                for m in range(3):
                                    dPhidu[nu, pi, i, j, k, l] = (
                                        fc3[p2s[nu], pi, :, i, j, :] *
                                        Y[:, :, k, l]).sum()
                                    # (Y[:,:,k,l] + Y[:,:,l,k]) / 2).sum() # Symmetrization?

        return dPhidu

    def _get_Y(self, nu):
        P = self._fc2
        X = self._X
        vecs = self._shortest_vectors
        multi = self._multiplicity
        lat = self._pcell.get_cell()
        num_atom_super = self._scell.get_number_of_atoms()
        R = np.array(
            [np.dot(vecs[Npi, nu, :multi[Npi,nu], :].sum(axis=0) /
                    multi[Npi,nu], lat) for Npi in range(num_atom_super)])

        p2s = self._pcell.get_primitive_to_supercell_map()
        s2p = self._pcell.get_supercell_to_primitive_map()
        p2p = self._pcell.get_primitive_to_primitive_map()

        Y = np.zeros((num_atom_super, 3, 3, 3), dtype=float)

        for Mmu in range(num_atom_super):
            for i in range(3):
                Y[Mmu, i, i, :] = R[Mmu, :]
            Y[Mmu] += X[p2p[s2p[Mmu]]]

        return Y

    def _get_X(self):
        num_atom_super = self._scell.get_number_of_atoms()
        num_atom_prim = self._pcell.get_number_of_atoms()
        p2s = self._pcell.get_primitive_to_supercell_map()
        lat = self._pcell.get_cell()
        vecs = self._shortest_vectors
        multi = self._multiplicity
        X = np.zeros((num_atom_prim, 3, 3, 3), dtype=float)
        G = self._get_Gamma()
        P = self._fc2

        for mu in range(num_atom_prim):
            for nu in range(num_atom_prim):
                R = np.array(
                    [np.dot(vecs[Npi, nu, :multi[Npi, nu], :].sum(axis=0) /
                            multi[Npi, nu], lat)
                     for Npi in range(num_atom_super)])
                for i in range(3):
                    for j in range(3):
                        for k in range(3):
                            for l in range(3):
                                X[mu, i, j, k] -= G[mu, nu, i, l] * \
                                    np.dot(P[p2s[nu], :, l, j], R[:, k])

        return X

    def _get_Gamma(self):
        num_atom_prim = self._pcell.get_number_of_atoms()
        m = self._pcell.get_masses()
        self._dm.set_dynamical_matrix([0, 0, 0])
        vals, vecs = np.linalg.eigh(self._dm.get_dynamical_matrix().real)
        G = np.zeros((num_atom_prim, num_atom_prim, 3, 3), dtype=float)

        for pi in range(num_atom_prim):
            for mu in range(num_atom_prim):
                for k in range(3):
                    for i in range(3):
                        # Eigenvectors are real.
                        # 3: means optical modes
                        G[pi, mu, k, i] = (
                            1.0 / np.sqrt(m[pi] * m[mu]) *
                            (vecs[pi * 3 + k, 3:] * vecs[mu * 3 + i, 3:] /
                             vals[3:]).sum())
        return G



import numpy as np

def get_phonons_at_qpoints(frequencies,
                           eigenvectors,
                           dm,
                           qpoints,
                           frequency_factor_to_THz,
                           nac_q_direction=None,
                           lapack_zheev_uplo='L'):
    import phono3py._lapackepy as lapackepy

    (svecs,
     multiplicity,
     masses,
     rec_lattice,
     born,
     nac_factor,
     dielectric) = _extract_params(dm)

    lapackepy.phonons_at_qpoints(
        frequencies,
        eigenvectors,
        np.array(qpoints, dtype='double', order='C'),
        dm.get_force_constants(),
        svecs,
        multiplicity,
        masses,
        dm.get_primitive_to_supercell_map(),
        dm.get_supercell_to_primitive_map(),
        frequency_factor_to_THz,
        born,
        dielectric,
        rec_lattice,
        nac_q_direction,
        nac_factor,
        lapack_zheev_uplo)

def set_phonon_c(dm,
                 frequencies,
                 eigenvectors,
                 phonon_done,
                 grid_points,
                 grid_address,
                 mesh,
                 frequency_factor_to_THz,
                 nac_q_direction,
                 lapack_zheev_uplo):
    import phono3py._lapackepy as lapackepy

    (svecs,
     multiplicity,
     masses,
     rec_lattice,
     born,
     nac_factor,
     dielectric) = _extract_params(dm)

    lapackepy.phonons_at_gridpoints(
        frequencies,
        eigenvectors,
        phonon_done,
        grid_points,
        grid_address,
        np.array(mesh, dtype='intc'),
        dm.get_force_constants(),
        svecs,
        multiplicity,
        masses,
        dm.get_primitive_to_supercell_map(),
        dm.get_supercell_to_primitive_map(),
        frequency_factor_to_THz,
        born,
        dielectric,
        rec_lattice,
        nac_q_direction,
        nac_factor,
        lapack_zheev_uplo)

def set_phonon_py(grid_point,
                  phonon_done,
                  frequencies,
                  eigenvectors,
                  grid_address,
                  mesh,
                  dynamical_matrix,
                  frequency_factor_to_THz,                  
                  lapack_zheev_uplo):
    gp = grid_point
    if phonon_done[gp] == 0:
        phonon_done[gp] = 1
        q = grid_address[gp].astype('double') / mesh
        dynamical_matrix.set_dynamical_matrix(q)
        dm = dynamical_matrix.get_dynamical_matrix()
        eigvals, eigvecs = np.linalg.eigh(dm, UPLO=lapack_zheev_uplo)
        eigvals = eigvals.real
        frequencies[gp] = (np.sqrt(np.abs(eigvals)) * np.sign(eigvals)
                           * frequency_factor_to_THz)
        eigenvectors[gp] = eigvecs

def _extract_params(dm):
    svecs, multiplicity = dm.get_shortest_vectors()
    masses = np.array(dm.get_primitive().get_masses(), dtype='double')
    rec_lattice = np.array(
        np.linalg.inv(dm.get_primitive().get_cell()), dtype='double', order='C')
    if dm.is_nac():
        born = dm.get_born_effective_charges()
        nac_factor = dm.get_nac_factor()
        dielectric = dm.get_dielectric_constant()
    else:
        born = None
        nac_factor = 0
        dielectric = None

    return (svecs,
            multiplicity,
            masses,
            rec_lattice,
            born,
            nac_factor,
            dielectric)



# Copyright (C) 2016 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from phono3py.phonon3.fc3 import distribute_fc3
from phonopy.harmonic.force_constants import distribute_force_constants

def get_fc2(supercell,
            forces_fc2,
            disp_dataset,
            symmetry):
    natom = supercell.get_number_of_atoms()
    force = np.array(forces_fc2, dtype='double', order='C')
    disp = np.zeros_like(force)
    lattice = supercell.get_cell()
    positions = supercell.get_scaled_positions()
    numbers = supercell.get_atomic_numbers()
    _set_disp_fc2(disp, disp_dataset)
    pure_trans = _collect_pure_translations(symmetry)
    rotations = np.array([np.eye(3, dtype='intc')] * len(pure_trans),
                         dtype='intc', order='C')

    print("------------------------------"
          " ALM FC2 start "
          "------------------------------")

    from alm import ALM
    with ALM(lattice, positions, numbers, 1) as alm:
        alm.set_displacement_and_force(disp, force)
        alm.run_fitting()
        fc2_alm = alm.get_fc(1)

    print("-------------------------------"
          " ALM FC2 end "
          "-------------------------------")

    fc2 = _expand_fc2(fc2_alm, supercell, pure_trans, rotations)

    return fc2

def get_fc3(supercell,
            forces_fc3,
            disp_dataset,
            symmetry):
    natom = supercell.get_number_of_atoms()
    force = np.array(forces_fc3, dtype='double', order='C')
    disp = np.zeros_like(force)
    lattice = supercell.get_cell()
    positions = supercell.get_scaled_positions()
    numbers = supercell.get_atomic_numbers()
    indices = _set_disp_fc3(disp, disp_dataset)
    pure_trans = _collect_pure_translations(symmetry)
    rotations = np.array([np.eye(3, dtype='intc')] * len(pure_trans),
                         dtype='intc', order='C')

    print("------------------------------"
          " ALM FC3 start "
          "------------------------------")

    from alm import ALM
    with ALM(lattice, positions, numbers, 2) as alm:
        alm.set_displacement_and_force(disp[indices], force[indices])
        if 'cutoff_distance' in disp_dataset:
            cut_d = disp_dataset['cutoff_distance']
            nkd = len(np.unique(numbers))
            rcs = np.ones((2, nkd, nkd), dtype='double')
            rcs[0] *= -1
            rcs[1] *= cut_d
            alm.set_cutoff_radii(rcs)
        alm.run_fitting()
        fc2_alm = alm.get_fc(1)
        fc3_alm = alm.get_fc(2)

    print("-------------------------------"
          " ALM FC3 end "
          "-------------------------------")

    fc2 = _expand_fc2(fc2_alm, supercell, pure_trans, rotations)
    fc3 = _expand_fc3(fc3_alm, supercell, pure_trans, rotations)

    return fc2, fc3

def _set_disp_fc2(disp, disp_dataset):
    count = 0
    for disp1 in disp_dataset['first_atoms']:
        disp[count, disp1['number']] = disp1['displacement']
        count += 1

def _set_disp_fc3(disp, disp_dataset):
    indices = []
    count = 0
    for disp1 in disp_dataset['first_atoms']:
        indices.append(count)
        disp[count, disp1['number']] = disp1['displacement']
        count += 1

    for disp1 in disp_dataset['first_atoms']:
        for disp2 in disp1['second_atoms']:
            if 'included' in disp2:
                if disp2['included']:
                    indices.append(count)
            else:
                indices.append(count)
            disp[count, disp1['number']] = disp1['displacement']
            disp[count, disp2['number']] = disp2['displacement']
            count += 1

    assert count == len(disp)

    return indices
                                       
def _expand_fc2(fc2_alm, supercell, pure_trans, rotations, symprec=1e-5):
    natom = supercell.get_number_of_atoms()
    fc2 = np.zeros((natom, natom, 3, 3), dtype='double', order='C')
    (fc_values, elem_indices) = fc2_alm
    first_atoms = np.unique(elem_indices[:, 0] // 3)

    for (fc, indices) in zip(fc_values, elem_indices):
        v1 = indices[0] // 3
        c1 = indices[0] % 3
        v2 = indices[1] // 3
        c2 = indices[1] % 3
        fc2[v1, v2, c1, c2] = fc

    lattice = np.array(supercell.get_cell().T, dtype='double', order='C')
    positions = supercell.get_scaled_positions()
    distribute_force_constants(fc2,
                               range(natom),
                               first_atoms,
                               lattice,
                               positions,
                               rotations,
                               pure_trans,
                               symprec)
    return fc2

def _expand_fc3(fc3_alm, supercell, pure_trans, rotations, symprec=1e-5):
    natom = supercell.get_number_of_atoms()
    fc3 = np.zeros((natom, natom, natom, 3, 3, 3), dtype='double', order='C')
    (fc_values, elem_indices) = fc3_alm
    first_atoms = np.unique(elem_indices[:, 0] // 3)

    for (fc, indices) in zip(fc_values, elem_indices):
        v1 = indices[0] // 3
        c1 = indices[0] % 3
        v2 = indices[1] // 3
        c2 = indices[1] % 3
        v3 = indices[2] // 3
        c3 = indices[2] % 3
        if v2 == v3 and c2 == c3:
            fc3[v1, v2, v3, c1, c2, c3] = fc
        else:
            fc3[v1, v2, v3, c1, c2, c3] = fc
            fc3[v1, v3, v2, c1, c3, c2] = fc

    lattice = np.array(supercell.get_cell().T, dtype='double', order='C')
    positions = supercell.get_scaled_positions()
    distribute_fc3(fc3,
                   first_atoms,
                   lattice,
                   positions,
                   rotations,
                   pure_trans,
                   symprec,
                   overwrite=True,
                   verbose=True)
    return fc3

def _collect_pure_translations(symmetry):
    pure_trans = []
    rotations = symmetry.get_symmetry_operations()['rotations']
    translations = symmetry.get_symmetry_operations()['translations']
    for r, t in zip(rotations, translations):
        if (r == np.eye(3, dtype='intc')).all():
            pure_trans.append(t)
    return np.array(pure_trans, dtype='double', order='C')



import numpy as np
from phonopy.units import THzToEv, Kb
import phonopy.structure.spglib as spg
from phonopy.structure.symmetry import Symmetry
from phonopy.structure.tetrahedron_method import TetrahedronMethod
from phonopy.structure.grid_points import extract_ir_grid_points

def gaussian(x, sigma):
    return 1.0 / np.sqrt(2 * np.pi) / sigma * np.exp(-x**2 / 2 / sigma**2)

def occupation(x, t):
    return 1.0 / (np.exp(THzToEv * x / (Kb * t)) - 1)


def get_triplets_at_q(grid_point,
                      mesh,
                      point_group, # real space point group of space group
                      primitive_lattice, # column vectors
                      is_time_reversal=True,
                      stores_triplets_map=False):
    map_triplets, map_q, grid_address = _get_triplets_reciprocal_mesh_at_q(
        grid_point,
        mesh,
        point_group,
        is_time_reversal=is_time_reversal)
    bz_grid_address, bz_map = spg.relocate_BZ_grid_address(grid_address,
                                                           mesh,
                                                           primitive_lattice)
    triplets_at_q, weights = _get_BZ_triplets_at_q(
        grid_point,
        bz_grid_address,
        bz_map,
        map_triplets,
        mesh)

    assert np.prod(mesh) == weights.sum(), \
        "Num grid points %d, sum of weight %d" % (
                    np.prod(mesh), weights.sum())

    # These maps are required for collision matrix calculation.
    if not stores_triplets_map:
        map_triplets = None
        map_q = None

    return triplets_at_q, weights, bz_grid_address, bz_map, map_triplets, map_q

def get_triplets_third_q_list(grid_point,
                              bz_grid_address,
                              bz_map,
                              mesh):
    triplets_at_q, weights = _get_BZ_triplets_at_q(
        grid_point,
        bz_grid_address,
        bz_map,
        np.arange(len(bz_grid_address), dtype='intc'),
        mesh)

    return np.array(triplets_at_q[:, 2], dtype='intc')

def get_nosym_triplets_at_q(grid_point,
                            mesh,
                            primitive_lattice,
                            stores_triplets_map=False):
    grid_address = get_grid_address(mesh)
    map_triplets = np.arange(len(grid_address), dtype='intc')
    bz_grid_address, bz_map = spg.relocate_BZ_grid_address(grid_address,
                                                           mesh,
                                                           primitive_lattice)
    triplets_at_q, weights = _get_BZ_triplets_at_q(
        grid_point,
        bz_grid_address,
        bz_map,
        map_triplets,
        mesh)

    if not stores_triplets_map:
        map_triplets = None
        map_q = None
    else:
        map_q = map_triplets.copy()

    return triplets_at_q, weights, bz_grid_address, bz_map, map_triplets, map_q


def get_grid_address(mesh):
    grid_mapping_table, grid_address = spg.get_stabilized_reciprocal_mesh(
        mesh,
        [[[1, 0, 0], [0, 1, 0], [0, 0, 1]]],
        is_time_reversal=False)

    return grid_address

def get_bz_grid_address(mesh, primitive_lattice, with_boundary=False):
    grid_address = get_grid_address(mesh)
    bz_grid_address, bz_map = spg.relocate_BZ_grid_address(grid_address,
                                                           mesh,
                                                           primitive_lattice)
    if with_boundary:
        return bz_grid_address, bz_map
    else:
        return bz_grid_address[:np.prod(mesh)]

def get_grid_point_from_address_py(address, mesh):
    # X runs first in XYZ
    # (*In spglib, Z first is possible with MACRO setting.)
    m = mesh
    return (address[0] % m[0] +
            (address[1] % m[1]) * m[0] +
            (address[2] % m[2]) * m[0] * m[1])

def get_grid_point_from_address(address, mesh):
    return spg.get_grid_point_from_address(address, mesh)

def get_bz_grid_point_from_address(address, mesh, bz_map):
    # X runs first in XYZ
    # (*In spglib, Z first is possible with MACRO setting.)
    # 2m is defined in kpoint.c of spglib.
    m = 2 * np.array(mesh, dtype='intc')
    return bz_map[get_grid_point_from_address(address, m)]

def invert_grid_point(grid_point, mesh, grid_address, bz_map):
    # gp --> [address] --> [-address] --> inv_gp
    address = grid_address[grid_point]
    return get_bz_grid_point_from_address(-address, mesh, bz_map)

def get_ir_grid_points(mesh, rotations, mesh_shifts=None):
    if mesh_shifts is None:
        mesh_shifts = [False, False, False]
    grid_mapping_table, grid_address = spg.get_stabilized_reciprocal_mesh(
        mesh,
        rotations,
        is_shift=np.where(mesh_shifts, 1, 0))
    (ir_grid_points,
     ir_grid_weights) = extract_ir_grid_points(grid_mapping_table)

    return ir_grid_points, ir_grid_weights, grid_address, grid_mapping_table

def get_grid_points_by_rotations(grid_point,
                                 reciprocal_rotations,
                                 mesh,
                                 mesh_shifts=None):
    if mesh_shifts is None:
        mesh_shifts = [False, False, False]
    return spg.get_grid_points_by_rotations(
        grid_point,
        reciprocal_rotations,
        mesh,
        is_shift=np.where(mesh_shifts, 1, 0))

def get_BZ_grid_points_by_rotations(grid_point,
                                    reciprocal_rotations,
                                    mesh,
                                    bz_map,
                                    mesh_shifts=None):
    if mesh_shifts is None:
        mesh_shifts = [False, False, False]
    return spg.get_BZ_grid_points_by_rotations(
        grid_point,
        reciprocal_rotations,
        mesh,
        bz_map,
        is_shift=np.where(mesh_shifts, 1, 0))

def reduce_grid_points(mesh_divisors,
                       grid_address,
                       dense_grid_points,
                       dense_grid_weights=None,
                       coarse_mesh_shifts=None):
    divisors = np.array(mesh_divisors, dtype='intc')
    if (divisors == 1).all():
        coarse_grid_points = np.array(dense_grid_points, dtype='intc')
        if dense_grid_weights is not None:
            coarse_grid_weights = np.array(dense_grid_weights, dtype='intc')
    else:
        grid_weights = []
        if coarse_mesh_shifts is None:
            shift = [0, 0, 0]
        else:
            shift = np.where(coarse_mesh_shifts, divisors // 2, [0, 0, 0])
        modulo = grid_address[dense_grid_points] % divisors
        condition = (modulo == shift).all(axis=1)
        coarse_grid_points = np.extract(condition, dense_grid_points)
        if dense_grid_weights is not None:
            coarse_grid_weights = np.extract(condition, dense_grid_weights)

    if dense_grid_weights is None:
        return coarse_grid_points
    else:
        return coarse_grid_points, coarse_grid_weights

def from_coarse_to_dense_grid_points(dense_mesh,
                                     mesh_divisors,
                                     coarse_grid_points,
                                     coarse_grid_address,
                                     coarse_mesh_shifts=None):
    if coarse_mesh_shifts is None:
        coarse_mesh_shifts = [False, False, False]
    shifts = np.where(coarse_mesh_shifts, 1, 0)
    dense_grid_points = []
    for cga in coarse_grid_address[coarse_grid_points]:
        dense_address = cga * mesh_divisors + shifts * (mesh_divisors // 2)
        dense_grid_points.append(get_grid_point_from_address(dense_address,
                                                             dense_mesh))
    return np.array(dense_grid_points, dtype='intc')

def get_coarse_ir_grid_points(primitive,
                              mesh,
                              mesh_divisors,
                              coarse_mesh_shifts,
                              is_kappa_star=True,
                              symprec=1e-5):
    mesh = np.array(mesh, dtype='intc')

    symmetry = Symmetry(primitive, symprec)
    point_group = symmetry.get_pointgroup_operations()

    if mesh_divisors is None:
        (ir_grid_points,
         ir_grid_weights,
         grid_address,
         grid_mapping_table) = get_ir_grid_points(mesh, point_group)
    else:
        mesh_divs = np.array(mesh_divisors, dtype='intc')
        coarse_mesh = mesh // mesh_divs
        if coarse_mesh_shifts is None:
            coarse_mesh_shifts = [False, False, False]

        if not is_kappa_star:
            coarse_grid_address = get_grid_address(coarse_mesh)
            coarse_grid_points = np.arange(np.prod(coarse_mesh), dtype='intc')
            coarse_grid_weights = np.ones(len(coarse_grid_points), dtype='intc')
        else:
            (coarse_ir_grid_points,
             coarse_ir_grid_weights,
             coarse_grid_address,
             coarse_grid_mapping_table) = get_ir_grid_points(
                 coarse_mesh,
                 point_group,
                 mesh_shifts=coarse_mesh_shifts)
        ir_grid_points = from_coarse_to_dense_grid_points(
            mesh,
            mesh_divs,
            coarse_grid_points,
            coarse_grid_address,
            coarse_mesh_shifts=coarse_mesh_shifts)
        grid_address = get_grid_address(mesh)
        ir_grid_weights = ir_grid_weights

    primitive_lattice = np.linalg.inv(primitive.get_cell())
    bz_grid_address, bz_map = spg.relocate_BZ_grid_address(grid_address,
                                                           mesh,
                                                           primitive_lattice)

    return (ir_grid_points,
            ir_grid_weights,
            bz_grid_address,
            grid_mapping_table)

def get_number_of_triplets(primitive,
                           mesh,
                           grid_point,
                           symprec=1e-5):
    mesh = np.array(mesh, dtype='intc')
    symmetry = Symmetry(primitive, symprec)
    point_group = symmetry.get_pointgroup_operations()
    primitive_lattice = np.linalg.inv(primitive.get_cell())
    triplets_at_q, _, _, _, _, _ = get_triplets_at_q(
        grid_point,
        mesh,
        point_group,
        primitive_lattice)

    return len(triplets_at_q)

def get_grid_points_in_Brillouin_zone(primitive_vectors, # column vectors
                                      mesh,
                                      grid_address,
                                      grid_points,
                                      with_boundary=False):
    gbz = GridBrillouinZone(primitive_vectors,
                            mesh,
                            grid_address,
                            with_boundary=with_boundary)
    gbz.run(grid_points)
    return gbz.get_shortest_addresses()

def get_triplets_integration_weights(interaction,
                                     frequency_points,
                                     sigma,
                                     sigma_cutoff=None,
                                     is_collision_matrix=False,
                                     neighboring_phonons=False,
                                     lang='C'):
    triplets = interaction.get_triplets_at_q()[0]
    frequencies = interaction.get_phonons()[0]
    num_band = frequencies.shape[1]
    g_zero = None

    if is_collision_matrix:
        g = np.empty(
            (3, len(triplets), len(frequency_points), num_band, num_band),
            dtype='double', order='C')
    else:
        g = np.empty(
            (2, len(triplets), len(frequency_points), num_band, num_band),
            dtype='double', order='C')
    g[:] = 0

    if sigma:
        if lang == 'C':
            import phono3py._phono3py as phono3c
            g_zero = np.zeros(g.shape[1:], dtype='byte', order='C')
            if sigma_cutoff is None:
                cutoff = -1
            else:
                cutoff = float(sigma_cutoff)
            # cutoff < 0 disables g_zero feature.
            phono3c.triplets_integration_weights_with_sigma(
                g,
                g_zero,
                frequency_points,
                triplets,
                frequencies,
                sigma,
                cutoff)
        else:
            for i, tp in enumerate(triplets):
                f1s = frequencies[tp[1]]
                f2s = frequencies[tp[2]]
                for j, k in list(np.ndindex((num_band, num_band))):
                    f1 = f1s[j]
                    f2 = f2s[k]
                    g0 = gaussian(frequency_points - f1 - f2, sigma)
                    g[0, i, :, j, k] = g0
                    g1 = gaussian(frequency_points + f1 - f2, sigma)
                    g2 = gaussian(frequency_points - f1 + f2, sigma)
                    g[1, i, :, j, k] = g1 - g2
                    if len(g) == 3:
                        g[2, i, :, j, k] = g0 + g1 + g2
    else:
        if lang == 'C':
            g_zero = np.zeros(g.shape[1:], dtype='byte', order='C')
            _set_triplets_integration_weights_c(
                g,
                g_zero,
                interaction,
                frequency_points,
                neighboring_phonons=neighboring_phonons)
        else:
            _set_triplets_integration_weights_py(
                g, interaction, frequency_points)

    return g, g_zero

def get_tetrahedra_vertices(relative_address,
                            mesh,
                            triplets_at_q,
                            bz_grid_address,
                            bz_map):
    bzmesh = mesh * 2
    grid_order = [1, mesh[0], mesh[0] * mesh[1]]
    bz_grid_order = [1, bzmesh[0], bzmesh[0] * bzmesh[1]]
    num_triplets = len(triplets_at_q)
    vertices = np.zeros((num_triplets, 2, 24, 4), dtype='intc')
    for i, tp in enumerate(triplets_at_q):
        for j, adrs_shift in enumerate(
                (relative_address, -relative_address)):
            adrs = bz_grid_address[tp[j + 1]] + adrs_shift
            bz_gp = np.dot(adrs % bzmesh, bz_grid_order)
            gp = np.dot(adrs % mesh, grid_order)
            vgp = bz_map[bz_gp]
            vertices[i, j] = vgp + (vgp == -1) * (gp + 1)
    return vertices

def _get_triplets_reciprocal_mesh_at_q(fixed_grid_number,
                                       mesh,
                                       rotations,
                                       is_time_reversal=True):
    import phono3py._phono3py as phono3c

    map_triplets = np.zeros(np.prod(mesh), dtype='intc')
    map_q = np.zeros(np.prod(mesh), dtype='intc')
    mesh_points = np.zeros((np.prod(mesh), 3), dtype='intc')

    phono3c.triplets_reciprocal_mesh_at_q(
        map_triplets,
        map_q,
        mesh_points,
        fixed_grid_number,
        np.array(mesh, dtype='intc'),
        is_time_reversal * 1,
        np.array(rotations, dtype='intc', order='C'))

    return map_triplets, map_q, mesh_points

def _get_BZ_triplets_at_q(grid_point,
                          bz_grid_address,
                          bz_map,
                          map_triplets,
                          mesh):
    """grid_address is overwritten."""

    import phono3py._phono3py as phono3c

    weights = np.zeros_like(map_triplets)
    for g in map_triplets:
        weights[g] += 1
    ir_weights = np.extract(weights > 0, weights)
    triplets = np.zeros((len(ir_weights), 3), dtype='intc')
    num_ir_ret = phono3c.BZ_triplets_at_q(triplets,
                                          grid_point,
                                          bz_grid_address,
                                          bz_map,
                                          map_triplets,
                                          np.array(mesh, dtype='intc'))

    return triplets, ir_weights


def _set_triplets_integration_weights_c(g,
                                        g_zero,
                                        interaction,
                                        frequency_points,
                                        neighboring_phonons=False):
    import phono3py._phono3py as phono3c

    reciprocal_lattice = np.linalg.inv(interaction.get_primitive().get_cell())
    mesh = interaction.get_mesh_numbers()
    thm = TetrahedronMethod(reciprocal_lattice, mesh=mesh)
    grid_address = interaction.get_grid_address()
    bz_map = interaction.get_bz_map()
    triplets_at_q = interaction.get_triplets_at_q()[0]

    if neighboring_phonons:
        unique_vertices = thm.get_unique_tetrahedra_vertices()
        for i, j in zip((1, 2), (1, -1)):
            neighboring_grid_points = np.zeros(
                len(unique_vertices) * len(triplets_at_q), dtype='intc')
            phono3c.neighboring_grid_points(
                neighboring_grid_points,
                triplets_at_q[:, i].flatten(),
                j * unique_vertices,
                mesh,
                grid_address,
                bz_map)
            interaction.set_phonons(np.unique(neighboring_grid_points))

    phono3c.triplets_integration_weights(
        g,
        g_zero,
        frequency_points,
        thm.get_tetrahedra(),
        mesh,
        triplets_at_q,
        interaction.get_phonons()[0],
        grid_address,
        bz_map)

def _set_triplets_integration_weights_py(g, interaction, frequency_points):
    reciprocal_lattice = np.linalg.inv(interaction.get_primitive().get_cell())
    mesh = interaction.get_mesh_numbers()
    thm = TetrahedronMethod(reciprocal_lattice, mesh=mesh)
    grid_address = interaction.get_grid_address()
    bz_map = interaction.get_bz_map()
    triplets_at_q = interaction.get_triplets_at_q()[0]
    unique_vertices = thm.get_unique_tetrahedra_vertices()
    tetrahedra_vertices = get_tetrahedra_vertices(
        thm.get_tetrahedra(),
        mesh,
        triplets_at_q,
        grid_address,
        bz_map)
    interaction.set_phonons(np.unique(tetrahedra_vertices))
    frequencies = interaction.get_phonons()[0]
    num_band = frequencies.shape[1]
    for i, vertices in enumerate(tetrahedra_vertices):
        for j, k in list(np.ndindex((num_band, num_band))):
            f1_v = frequencies[vertices[0], j]
            f2_v = frequencies[vertices[1], k]
            thm.set_tetrahedra_omegas(f1_v + f2_v)
            thm.run(frequency_points)
            g0 = thm.get_integration_weight()
            g[0, i, :, j, k] = g0
            thm.set_tetrahedra_omegas(-f1_v + f2_v)
            thm.run(frequency_points)
            g1 = thm.get_integration_weight()
            thm.set_tetrahedra_omegas(f1_v - f2_v)
            thm.run(frequency_points)
            g2 = thm.get_integration_weight()
            g[1, i, :, j, k] = g1 - g2
            if len(g) == 3:
                g[2, i, :, j, k] = g0 + g1 + g2

class GridBrillouinZone(object):
    search_space = np.array([
            [0, 0, 0],
            [0, 0, 1],
            [0, 1, -1],
            [0, 1, 0],
            [0, 1, 1],
            [1, -1, -1],
            [1, -1, 0],
            [1, -1, 1],
            [1, 0, -1],
            [1, 0, 0],
            [1, 0, 1],
            [1, 1, -1],
            [1, 1, 0],
            [1, 1, 1],
            [-1, -1, -1],
            [-1, -1, 0],
            [-1, -1, 1],
            [-1, 0, -1],
            [-1, 0, 0],
            [-1, 0, 1],
            [-1, 1, -1],
            [-1, 1, 0],
            [-1, 1, 1],
            [0, -1, -1],
            [0, -1, 0],
            [0, -1, 1],
            [0, 0, -1]], dtype='intc', order='C')

    def __init__(self,
                 primitive_vectors,
                 mesh,
                 grid_address,
                 with_boundary=False): # extended grid if True
        self._primitive_vectors = np.array(primitive_vectors) # column vectors
        self._mesh = mesh
        self._grid_address = grid_address
        self._with_boundary = with_boundary

        self._tolerance = min(np.sum(self._primitive_vectors ** 2, axis=0)) / 10
        self._primitive_vectors_inv = np.linalg.inv(self._primitive_vectors)
        self._search_space = search_space * mesh

        self._shortest_addresses = None

    def run(self, grid_points):
        self._shortest_addresses = []
        for address in self._grid_address[grid_points]:
            distances = np.array(
                [(np.dot(self._primitive_vectors, address + g) ** 2).sum()
                 for g in self._search_space], dtype='double')
            min_dist = min(distances)
            shortest_indices = [i for i, d in enumerate(distances - min_dist)
                                if abs(d) < self._tolerance]
            self._shortest_addresses.append(
                self._search_space[shortest_indices] + address)

    def get_shortest_addresses(self):
        return self._shortest_addresses



# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
from phonopy.cui.phonopy_argparse import fix_deprecated_option_names

def get_parser():
    deprecated = fix_deprecated_option_names(sys.argv)

    import argparse
    parser = argparse.ArgumentParser(
        description="Phono3py command-line-tool")
    parser.set_defaults(abinit_mode=False,
                        band_indices=None,
                        band_paths=None,
                        band_points=None,
                        cell_filename=None,
                        constant_averaged_pp_interaction=None,
                        crystal_mode=False,
                        cutoff_fc3_distance=None,
                        cutoff_frequency=None,
                        boundary_mfp=None,
                        cutoff_pair_distance=None,
                        delta_fc2=False,
                        delta_fc2_sets_mode=False,
                        displacement_distance=None,
                        force_sets_to_forces_fc2_mode=None,
                        forces_fc3=False,
                        forces_fc3_file=False,
                        forces_fc2_mode=False,
                        force_sets_mode=False,
                        frequency_conversion_factor=None,
                        fpitch=None,
                        frequency_scale_factor=None,
                        num_frequency_points=None,
                        freq_scale=None,
                        gamma_unit_conversion=None,
                        grid_addresses=None,
                        grid_points=None,
                        gv_delta_q=None,
                        input_filename=None,
                        input_output_filename=None,
                        ion_clamped=False,
                        is_bterta=False,
                        is_decay_channel=False,
                        is_displacement=False,
                        is_frequency_shift=False,
                        is_full_pp=False,
                        is_gruneisen=False,
                        is_isotope=False,
                        is_joint_dos=False,
                        is_linewidth=False,
                        is_lbte=False,
                        is_nac=False,
                        is_nodiag=False,
                        is_nomeshsym=False,
                        is_nosym=False,
                        is_N_U=False,
                        is_plusminus_displacements=False,
                        is_reducible_collision_matrix=False,
                        is_translational_symmetry=False,
                        is_symmetrize_fc2=False,
                        is_symmetrize_fc3_r=False,
                        is_symmetrize_fc3_q=False,
                        is_tetrahedron_method=False,
                        log_level=None,
                        max_freepath=None,
                        masses=None,
                        mass_variances=None,
                        mesh_numbers=None,
                        mesh_divisors=None,
                        no_kappa_stars=False,
                        output_filename=None,
                        phonon_supercell_dimension=None,
                        pinv_cutoff=None,
                        pinv_solver=None,
                        pp_unit_conversion=None,
                        primitive_axis=None,
                        qpoints=None,
                        quiet=False,
                        q_direction=None,
                        read_amplitude=False,
                        read_collision=None,
                        read_fc2=False,
                        read_fc3=False,
                        read_gamma=False,
                        scattering_event_class=None,
                        show_num_triplets=False,
                        sigma=None,
                        sigma_cutoff=None,
                        supercell_dimension=None,
                        symprec=1e-5,
                        temperatures=None,
                        tmax=None,
                        tmin=None,
                        tstep=None,
                        tsym_type=None,
                        uplo='L',
                        use_alm=False,
                        use_ave_pp=False,
                        verbose=False,
                        write_amplitude=False,
                        write_collision=False,
                        write_gamma_detail=False,
                        write_gamma=False,
                        write_phonon=False,
                        write_grid_points=False)
    parser.add_argument(
        "--abinit", dest="abinit_mode", action="store_true",
        help="Invoke Abinit mode")
    parser.add_argument(
        "--alm", dest="use_alm", action="store_true",
        help="Use ALM for creating force constants")
    parser.add_argument(
        "--amplitude", dest="displacement_distance", type=float,
        help="Distance of displacements")
    parser.add_argument(
        "--band", dest="band_paths",
        help="Band structure paths calculated for Gruneisen parameter")
    parser.add_argument(
        "--band-points", dest="band_points", type=int,
        help=("Number of points calculated on a band segment in the band "
              "structure Gruneisen parameter calculation"))
    parser.add_argument(
        "--bi", "--band-indices", dest="band_indices",
        help="Band indices where life time is calculated")
    parser.add_argument(
        "--boundary-mfp", "--bmfp", dest="boundary_mfp", type=float,
        help=("Boundary mean free path in micrometre for thermal conductivity "
              "calculation"))
    parser.add_argument(
        "--br", "--bterta", dest="is_bterta", action="store_true",
        help="Calculate thermal conductivity in BTE-RTA")
    parser.add_argument(
        "-c", "--cell", dest="cell_filename", metavar="FILE",
        help="Read unit cell")
    parser.add_argument(
        "--cf2", "--create-f2", dest="forces_fc2", nargs='+',
        help="Create FORCES_FC2")
    parser.add_argument(
        "--cf3", "--create-f3", dest="forces_fc3", nargs='+',
        help="Create FORCES_FC3")
    parser.add_argument(
        "--cf3-file", "--create-f3-from-file", dest="forces_fc3_file", nargs=1,
        help="Create FORCES_FC3 from file name list")
    parser.add_argument(
        "--cfs", "--create-force-sets", dest="force_sets_mode",
        action="store_true",
        help="Create phonopy FORCE_SETS from FORCES_FC2")
    parser.add_argument(
        "--const-ave-pp", dest="constant_averaged_pp_interaction", type=float,
        help="Set constant averaged ph-ph interaction (Pqj)")
    parser.add_argument(
        "--crystal", dest="crystal_mode", action="store_true",
        help="Invoke CRYSTAL mode")
    parser.add_argument(
        "--cutoff-fc3", "--cutoff-fc3-distance", dest="cutoff_fc3_distance",
        type=float,
        help=("Cutoff distance of third-order force constants. Elements where "
              "any pair of atoms has larger distance than cut-off distance are "
              "set zero."))
    parser.add_argument(
        "--cutoff-freq", "--cutoff-frequency", dest="cutoff_frequency",
        type=float,
        help="Phonon modes below this frequency are ignored.")
    parser.add_argument(
        "--cutoff-pair", "--cutoff-pair-distance", dest="cutoff_pair_distance",
        type=float,
        help=("Cutoff distance between pairs of displaced atoms used for "
              "supercell creation with displacements and making third-order "
              "force constants"))
    parser.add_argument(
        "-d", "--disp", dest="is_displacement", action="store_true",
        help="As first stage, get least displacements")
    parser.add_argument(
        "--dim", dest="supercell_dimension",
        help="Supercell dimension")
    parser.add_argument(
        "--dim-fc2", dest="phonon_supercell_dimension",
        help="Supercell dimension for extra fc2")
    parser.add_argument(
        "--factor", dest="frequency_conversion_factor", type=float,
        help="Conversion factor to favorite frequency unit")
    parser.add_argument(
        "--fc2", dest="read_fc2", action="store_true",
        help="Read second order force constants")
    parser.add_argument(
        "--fc3", dest="read_fc3", action="store_true",
        help="Read third order force constants")
    parser.add_argument(
        "--freq-scale", dest="frequency_scale_factor", type=float,
        help=("Squared scale factor multiplied with fc2. Therefore frequency "
              "is changed but the contribution from NAC is not changed."))
    parser.add_argument(
        "--freq-pitch", dest="fpitch", type=float,
        help="Pitch in frequency for spectrum")
    parser.add_argument(
        "--fs2f2", "--force-sets-to-forces-fc2",
        dest="force_sets_to_forces_fc2_mode",
        action="store_true", help="Create FORCES_FC2 from FORCE_SETS")
    parser.add_argument(
        "--fst", "--frequency-shift", dest="is_frequency_shift",
        action="store_true", help="Calculate frequency shifts")
    parser.add_argument(
        "--full-pp", dest="is_full_pp", action="store_true",
        help=("Calculate full ph-ph interaction for RTA conductivity."
              "This may be activated when full elements of ph-ph interaction "
              "strength are needed, i.e., to calculate average ph-ph "
              "interaction strength."))
    parser.add_argument(
        "--ga", "--grid-addresses", dest="grid_addresses",
        help="Fixed grid addresses where anharmonic properties are calculated")
    parser.add_argument(
        "--gamma-unit-conversion", dest="gamma_unit_conversion", type=float,
        help="Conversion factor for gamma")
    parser.add_argument(
        "--gp", "--grid-points", dest="grid_points",
        help="Fixed grid points where anharmonic properties are calculated")
    parser.add_argument(
        "--gruneisen", dest="is_gruneisen", action="store_true",
        help="Calculate phonon Gruneisen parameter")
    parser.add_argument(
        "--gv-delta-q", dest="gv_delta_q", type=float,
        help="Delta-q distance used for group velocity calculation")
    parser.add_argument(
        "-i", dest="input_filename",
        help="Input filename extension")
    parser.add_argument(
        "--io", dest="input_output_filename",
        help="Input and output filename extension")
    parser.add_argument(
        "--ion-clamped", dest="ion_clamped", action="store_true",
        help=("Atoms are clamped under applied strain in Gruneisen parameter "
              "calculation"))
    parser.add_argument(
        "--ise", dest="is_imag_self_energy", action="store_true",
        help="Calculate imaginary part of self energy")
    parser.add_argument(
        "--isotope", dest="is_isotope", action="store_true",
        help="Isotope scattering lifetime")
    parser.add_argument(
        "--jdos", dest="is_joint_dos", action="store_true",
        help="Calculate joint density of states")
    parser.add_argument(
        "--lbte", dest="is_lbte", action="store_true",
        help="Calculate thermal conductivity LBTE with Chaput's method")
    parser.add_argument(
        "--loglevel", dest="log_level", type=int,
        help="Log level")
    parser.add_argument(
        "--lw", "--linewidth", dest="is_linewidth", action="store_true",
        help="Calculate linewidths")
    parser.add_argument(
        "--mass", dest="masses",
        help="Same as MASS tag")
    parser.add_argument(
        "--md", "--mesh-divisors", dest="mesh_divisors",
        help="Divisors for mesh numbers")
    parser.add_argument(
        "--mesh", dest="mesh_numbers",
        help="Mesh numbers")
    parser.add_argument(
        "--mv", "--mass-variances", dest="mass_variances",
        help="Mass variance parameters for isotope scattering")
    parser.add_argument(
        "--nac", dest="is_nac", action="store_true",
        help="Non-analytical term correction")
    parser.add_argument(
        "--nodiag", dest="is_nodiag", action="store_true",
        help="Set displacements parallel to axes")
    parser.add_argument(
        "--noks", "--no-kappa-stars", dest="no_kappa_stars",
        action="store_true",
        help="Deactivate summation of partial kappa at q-stars"),
    parser.add_argument(
        "--nomeshsym", dest="is_nomeshsym", action="store_true",
        help="No symmetrization of triplets is made.")
    parser.add_argument(
        "--nosym", dest="is_nosym", action="store_true",
        help="Symmetry is not imposed.")
    parser.add_argument(
        "--nu", dest="is_N_U", action="store_true",
        help="Split Gamma into Normal and Umklapp processes")
    parser.add_argument(
        "--num-freq-points", dest="num_frequency_points", type=int,
        help="Number of sampling points for spectrum")
    parser.add_argument(
        "-o", dest="output_filename",
        help="Output filename extension")
    parser.add_argument(
        "--pa", "--primitive-axis", dest="primitive_axis",
        help="Same as PRIMITIVE_AXIS tags")
    parser.add_argument(
        "--pinv-cutoff", dest="pinv_cutoff", type=float,
        help="Cutoff frequency (THz) for pseudo inversion of collision matrix")
    parser.add_argument(
        "--pinv-solver", dest="pinv_solver", type=int,
        help="Switch of LBTE pinv solver")
    parser.add_argument(
        "--pm", dest="is_plusminus_displacements", action="store_true",
        help="Set plus minus displacements")
    parser.add_argument(
        "--pp-unit-conversion", dest="pp_unit_conversion", type=float,
        help="Conversion factor for ph-ph interaction")
    parser.add_argument(
        "--pwscf", dest="pwscf_mode", action="store_true",
        help="Invoke Pwscf mode")
    parser.add_argument(
        "--qpoints", dest="qpoints",
        help="Calculate at specified q-points")
    parser.add_argument(
        "--q-direction", dest="q_direction",
        help="q-vector direction at q->0 for non-analytical term correction")
    parser.add_argument(
        "-q", "--quiet", dest="quiet", action="store_true",
        help="Print out smallest information")
    # parser.add_argument(
    #     "--read_amplitude", dest="read_amplitude", action="store_true",
    #     help="Read phonon-phonon interaction amplitudes")
    parser.add_argument(
        "--ave-pp", dest="use_ave_pp", action="store_true",
        help="Use averaged ph-ph interaction")
    parser.add_argument(
        "--read-collision", dest="read_collision",
        help="Read collision matrix and Gammas from files")
    parser.add_argument(
        "--read-gamma", dest="read_gamma", action="store_true",
        help="Read Gammas from files")
    parser.add_argument(
        "--read-phonon", dest="read_phonon", action="store_true",
        help="Read phonons from files")
    parser.add_argument(
        "--reducible-colmat", dest="is_reducible_collision_matrix",
        action="store_true",
        help="Solve reducible collision matrix")
    parser.add_argument(
        "--scattering-event-class", dest="scattering_event_class", type=int,
        help=("Scattering event class 1 or 2 to draw imaginary part of self "
              "energy"))
    parser.add_argument(
        "--stp", "--show-num-triplets", dest="show_num_triplets",
        action="store_true",
        help=("Show reduced number of triplets to be calculated at "
              "specified grid points"))
    parser.add_argument(
        "--sigma", dest="sigma",
        help=("A sigma value or multiple sigma values (separated by space) for "
              "smearing function"))
    parser.add_argument(
        "--sigma-cutoff", dest="sigma_cutoff_width", type=float,
        help="Cutoff width of smearing function (ratio to sigma value)")
    parser.add_argument(
        "--sym-fc2", dest="is_symmetrize_fc2", action="store_true",
        help="Symmetrize fc2 by index exchange")
    parser.add_argument(
        "--sym-fc3r", dest="is_symmetrize_fc3_r", action="store_true",
        help="Symmetrize fc3 in real space by index exchange")
    parser.add_argument(
        "--sym-fc3q", dest="is_symmetrize_fc3_q", action="store_true",
        help="Symmetrize fc3 in reciprocal space by index exchange")
    parser.add_argument(
        "--thm", "--tetrahedron-method", dest="is_tetrahedron_method",
        action="store_true",
        help="Use tetrahedron method")
    parser.add_argument(
        "--tmax", dest="tmax",
        help="Maximum calculated temperature")
    parser.add_argument(
        "--tmin", dest="tmin",
        help="Minimum calculated temperature")
    parser.add_argument(
        "--ts", dest="temperatures",
        help="Temperatures for damping functions")
    parser.add_argument(
        "--tstep", dest="tstep",
        help="Calculated temperature step")
    parser.add_argument(
        "--tsym", dest="is_translational_symmetry", action="store_true",
        help="Impose translational invariance condition")
    parser.add_argument(
        "--tsym-type", dest="tsym_type", type=int,
        help="Imposing type of translational invariance")
    parser.add_argument(
        "--tolerance", dest="symprec", type=float,
        help="Symmetry tolerance to search")
    parser.add_argument(
        "--uplo", dest="uplo",
        help="Lapack zheev UPLO")
    parser.add_argument(
        "-v", "--verbose", dest="verbose", action="store_true",
        help="Detailed run-time information is displayed")
    parser.add_argument(
        "--wgp", "--write-grid-points", dest="write_grid_points",
        action="store_true",
        help=("Write grid address of irreducible grid points for specified "
              "mesh numbers to ir_grid_address.yaml"))
    # parser.add_argument("--write_amplitude", dest="write_amplitude",
    #                   action="store_true",
    #                   help="Write phonon-phonon interaction amplitudes")
    parser.add_argument(
        "--write-collision", dest="write_collision", action="store_true",
        help="Write collision matrix and Gammas to files")
    parser.add_argument(
        "--write-gamma", dest="write_gamma", action="store_true",
        help="Write imag-part of self energy to files")
    parser.add_argument(
        "--write-gamma-detail", "--write_detailed_gamma",
        dest="write_gamma_detail", action="store_true",
        help="Write out detailed imag-part of self energy")
    parser.add_argument(
        "--write-phonon", dest="write_phonon", action="store_true",
        help="Write all phonons on grid points to files")
    parser.add_argument(
        "conf_file", nargs='*',
        help="Phono3py configure file")

    return parser, deprecated



import os
import numpy as np
import h5py

def write_cell_yaml(w, supercell):
    w.write("lattice:\n")
    for axis in supercell.get_cell():
        w.write("- [ %20.15f,%20.15f,%20.15f ]\n" % tuple(axis))
    symbols = supercell.get_chemical_symbols()
    positions = supercell.get_scaled_positions()
    w.write("atoms:\n")
    for i, (s, v) in enumerate(zip(symbols, positions)):
        w.write("- symbol: %-2s # %d\n" % (s, i+1))
        w.write("  position: [ %18.14f,%18.14f,%18.14f ]\n" % tuple(v))

def write_disp_fc4_yaml(dataset, supercell, filename='disp_fc4.yaml'):
    w = open(filename, 'w')
    w.write("natom: %d\n" %  dataset['natom'])

    num_first = len(dataset['first_atoms'])
    w.write("num_first_displacements: %d\n" %  num_first)

    num_second = 0
    for d1 in dataset['first_atoms']:
        num_second += len(d1['second_atoms'])
    w.write("num_second_displacements: %d\n" %  num_second)

    num_third = 0
    for d1 in dataset['first_atoms']:
        for d2 in d1['second_atoms']:
            num_third += len(d2['third_atoms'])
    w.write("num_third_displacements: %d\n" %  num_third)

    w.write("first_atoms:\n")
    count1 = 1
    count2 = num_first + 1
    count3 = num_first + num_second + 1
    for disp1 in dataset['first_atoms']:
        disp_cart1 = disp1['displacement']
        w.write("- number: %5d\n" % (disp1['number'] + 1))
        w.write("  displacement:\n")
        w.write("    [%20.16f,%20.16f,%20.16f ] # %05d\n" %
                (disp_cart1[0], disp_cart1[1], disp_cart1[2], count1))
        w.write("  second_atoms:\n")
        count1 += 1
        for disp2 in disp1['second_atoms']:
            w.write("  - number: %5d\n" % (disp2['number'] + 1))
            w.write("    displacement:\n")
            disp_cart2 = disp2['displacement']
            w.write("      [%20.16f,%20.16f,%20.16f ] # %05d\n" %
                    (disp_cart2[0], disp_cart2[1], disp_cart2[2], count2))
            w.write("    third_atoms:\n")
            count2 += 1
            atom3 = -1
            for disp3 in disp2['third_atoms']:
                if atom3 != disp3['number']:
                    atom3 = disp3['number']
                    w.write("    - number: %5d\n" % (atom3 + 1))
                    w.write("      displacements:\n")
                disp_cart3 = disp3['displacement']
                w.write("      - [%20.16f,%20.16f,%20.16f ] # %05d\n" %
                        (disp_cart3[0], disp_cart3[1], disp_cart3[2], count3))
                count3 += 1

    write_cell_yaml(w, supercell)

    w.close()

    return num_first + num_second + num_third

def write_disp_fc3_yaml(dataset, supercell, filename='disp_fc3.yaml'):
    w = open(filename, 'w')
    w.write("natom: %d\n" %  dataset['natom'])

    num_first = len(dataset['first_atoms'])
    w.write("num_first_displacements: %d\n" %  num_first)
    if 'cutoff_distance' in dataset:
        w.write("cutoff_distance: %f\n" %  dataset['cutoff_distance'])

    num_second = 0
    num_disp_files = 0
    for d1 in dataset['first_atoms']:
        num_disp_files += 1
        num_second += len(d1['second_atoms'])
        for d2 in d1['second_atoms']:
            if 'included' in d2:
                if d2['included']:
                    num_disp_files += 1
            else:
                num_disp_files += 1

    w.write("num_second_displacements: %d\n" %  num_second)
    w.write("num_displacements_created: %d\n" %  num_disp_files)

    w.write("first_atoms:\n")
    count1 = 1
    count2 = num_first + 1
    for disp1 in dataset['first_atoms']:
        disp_cart1 = disp1['displacement']
        w.write("- number: %5d\n" % (disp1['number'] + 1))
        w.write("  displacement:\n")
        w.write("    [%20.16f,%20.16f,%20.16f ] # %05d\n" %
                (disp_cart1[0], disp_cart1[1], disp_cart1[2], count1))
        w.write("  second_atoms:\n")
        count1 += 1

        included = None
        distance = 0.0
        atom2 = -1
        for disp2 in disp1['second_atoms']:
            if atom2 != disp2['number']:
                atom2 = disp2['number']
                if 'included' in disp2:
                    included = disp2['included']
                pair_distance = disp2['pair_distance']
                w.write("  - number: %5d\n" % (atom2 + 1))
                w.write("    distance: %f\n" % pair_distance)
                if included is not None:
                    if included:
                        w.write("    included: %s\n" % "true")
                    else:
                        w.write("    included: %s\n" % "false")
                w.write("    displacements:\n")

            disp_cart2 = disp2['displacement']
            w.write("    - [%20.16f,%20.16f,%20.16f ] # %05d\n" %
                    (disp_cart2[0], disp_cart2[1], disp_cart2[2], count2))
            count2 += 1

    write_cell_yaml(w, supercell)

    w.close()

    return num_first + num_second, num_disp_files

def write_disp_fc2_yaml(dataset, supercell, filename='disp_fc2.yaml'):
    w = open(filename, 'w')
    w.write("natom: %d\n" %  dataset['natom'])

    num_first = len(dataset['first_atoms'])
    w.write("num_first_displacements: %d\n" %  num_first)
    w.write("first_atoms:\n")
    for i, disp1 in enumerate(dataset['first_atoms']):
        disp_cart1 = disp1['displacement']
        w.write("- number: %5d\n" % (disp1['number'] + 1))
        w.write("  displacement:\n")
        w.write("    [%20.16f,%20.16f,%20.16f ] # %05d\n" %
                (disp_cart1[0], disp_cart1[1], disp_cart1[2], i + 1))

    if supercell is not None:
        write_cell_yaml(w, supercell)

    w.close()

    return num_first

def write_FORCES_FC2(disp_dataset,
                     forces_fc2=None,
                     fp=None,
                     filename="FORCES_FC2"):
    if fp is None:
        w = open(filename, 'w')
    else:
        w = fp

    for i, disp1 in enumerate(disp_dataset['first_atoms']):
        w.write("# File: %-5d\n" % (i + 1))
        w.write("# %-5d " % (disp1['number'] + 1))
        w.write("%20.16f %20.16f %20.16f\n" % tuple(disp1['displacement']))
        if forces_fc2 is None:
            force_set = disp1['forces']
        else:
            force_set = forces_fc2[i]
        for forces in force_set:
            w.write("%15.10f %15.10f %15.10f\n" % tuple(forces))

def write_FORCES_FC3(disp_dataset, forces_fc3, fp=None, filename="FORCES_FC3"):
    if fp is None:
        w = open(filename, 'w')
    else:
        w = fp

    natom = disp_dataset['natom']
    num_disp1 = len(disp_dataset['first_atoms'])
    count = num_disp1
    file_count = num_disp1

    write_FORCES_FC2(disp_dataset, forces_fc2=forces_fc3, fp=w)

    for i, disp1 in enumerate(disp_dataset['first_atoms']):
        atom1 = disp1['number']
        for disp2 in disp1['second_atoms']:
            atom2 = disp2['number']
            w.write("# File: %-5d\n" % (count + 1))
            w.write("# %-5d " % (atom1 + 1))
            w.write("%20.16f %20.16f %20.16f\n" % tuple(disp1['displacement']))
            w.write("# %-5d " % (atom2 + 1))
            w.write("%20.16f %20.16f %20.16f\n" % tuple(disp2['displacement']))

            # For supercell calculation reduction
            included = True
            if 'included' in disp2:
                included = disp2['included']
            if included:
                for forces in forces_fc3[file_count]:
                    w.write("%15.10f %15.10f %15.10f\n" % tuple(forces))
                file_count += 1
            else:
                # for forces in forces_fc3[i]:
                #     w.write("%15.10f %15.10f %15.10f\n" % (tuple(forces)))
                for j in range(natom):
                    w.write("%15.10f %15.10f %15.10f\n" % (0, 0, 0))
            count += 1

def write_FORCES_FC4(disp_dataset, forces_fc4, fp=None, filename="FORCES_FC4"):
    if fp is None:
        w = open(filename, 'w')
    else:
        w = fp

    natom = disp_dataset['natom']
    num_disp1 = len(disp_dataset['first_atoms'])
    num_disp2 = 0
    for disp1 in disp_dataset['first_atoms']:
        num_disp2 += len(disp1['second_atoms'])
    count = num_disp1 + num_disp2

    write_FORCES_FC3(disp_dataset, forces_fc3=forces_fc4, fp=w)

    for i, disp1 in enumerate(disp_dataset['first_atoms']):
        atom1 = disp1['number']
        for disp2 in disp1['second_atoms']:
            atom2 = disp2['number']
            for disp3 in disp2['third_atoms']:
                atom3 = disp3['number']
                w.write("# File: %-5d\n" % (count + 1))
                w.write("# %-5d " % (atom1 + 1))
                w.write("%20.16f %20.16f %20.16f\n" %
                        tuple(disp1['displacement']))
                w.write("# %-5d " % (atom2 + 1))
                w.write("%20.16f %20.16f %20.16f\n" %
                        tuple(disp2['displacement']))
                w.write("# %-5d " % (atom3 + 1))
                w.write("%20.16f %20.16f %20.16f\n" %
                        tuple(disp3['displacement']))
                for forces in forces_fc4[count]:
                    w.write("%15.10f %15.10f %15.10f\n" % tuple(forces))
                count += 1

def write_fc3_yaml(force_constants_third,
                   filename='fc3.yaml',
                   is_symmetrize=False):
    w = open(filename, 'w')
    num_atom = force_constants_third.shape[0]
    for i in range(num_atom):
        for j in range(num_atom):
            for k in range(num_atom):
                if is_symmetrize:
                    tensor3 = symmetrize_fc3(force_constants_third, i, j, k)
                else:
                    tensor3 = force_constants_third[i, j, k]
                w.write("- index: [ %d - %d - %d ] # (%f)\n" \
                               % (i+1, j+1, k+1, tensor3.sum()))
                w.write("  tensor:\n")
                for tensor2 in tensor3:
                    w.write("  -\n")
                    for vec in tensor2:
                        w.write("    - [ %13.8f, %13.8f, %13.8f ]\n" % tuple(vec))
                w.write("\n")

def write_fc3_dat(force_constants_third, filename='fc3.dat'):
    w = open(filename, 'w')
    for i in range(force_constants_third.shape[0]):
        for j in range(force_constants_third.shape[1]):
            for k in range(force_constants_third.shape[2]):
                tensor3 = force_constants_third[i, j, k]
                w.write(" %d - %d - %d  (%f)\n" % (i + 1, j + 1, k + 1,
                                                   np.abs(tensor3).sum()))
                for tensor2 in tensor3:
                    for vec in tensor2:
                        w.write("%20.14f %20.14f %20.14f\n" % tuple(vec))
                    w.write("\n")

def write_fc4_dat(fc4, filename='fc4.dat'):
    w = open(filename, 'w')
    for (i, j, k, l) in list(np.ndindex(fc4.shape[:4])):
        tensor4 = fc4[i, j, k, l]
        w.write(" %d - %d - %d - %d (%f)\n" % (i + 1, j + 1, k + 1, l + 1,
                                               np.abs(tensor4).sum()))
        for tensor3 in tensor4:
            for tensor2 in tensor3:
                for vec in tensor2:
                    w.write("%20.14f %20.14f %20.14f\n" % tuple(vec))
                w.write("\n")
            w.write("\n")
        w.write("\n")

def write_fc4_to_hdf5(force_constants_fourth, filename='fc4.hdf5'):
    w = h5py.File(filename, 'w')
    w.create_dataset('fc4', data=force_constants_fourth)
    w.close()

def read_fc4_from_hdf5(filename='fc4.hdf5'):
    f = h5py.File(filename, 'r')
    fc4 = f['fc4'][:]
    f.close()
    return fc4

def write_fc3_to_hdf5(force_constants_third, filename='fc3.hdf5'):
    with h5py.File(filename, 'w') as w:
        w.create_dataset('fc3', data=force_constants_third)

def read_fc3_from_hdf5(filename='fc3.hdf5'):
    with h5py.File(filename, 'r') as f:
        fc3 = f['fc3'][:]
        return fc3
    return None

def write_fc2_dat(force_constants, filename='fc2.dat'):
    w = open(filename, 'w')
    for i, fcs in enumerate(force_constants):
        for j, fcb in enumerate(fcs):
            w.write(" %d - %d\n" % (i+1, j+1))
            for vec in fcb:
                w.write("%20.14f %20.14f %20.14f\n" % tuple(vec))
            w.write("\n")

def write_fc2_to_hdf5(force_constants, filename='fc2.hdf5'):
    with h5py.File(filename, 'w') as w:
        w.create_dataset('fc2', data=force_constants)

def read_fc2_from_hdf5(filename='fc2.hdf5'):
    with h5py.File(filename, 'r') as f:
        if 'fc2' in f.keys():
            fc2 = f['fc2'][:]
        elif 'force_constants' in f.keys():
            fc2 = f['force_constants'][:]
        else:
            fc2 = None
        return fc2
    return None

def write_triplets(triplets,
                   weights,
                   mesh,
                   grid_address,
                   grid_point=None,
                   filename=None):
    triplets_filename = "triplets"
    suffix = "-m%d%d%d" % tuple(mesh)
    if grid_point is not None:
        suffix += ("-g%d" % grid_point)
    if filename is not None:
        suffix += "." + filename
    suffix += ".dat"
    triplets_filename += suffix
    w = open(triplets_filename, 'w')
    for weight, g3 in zip(weights, triplets):
        w.write("%4d    " % weight)
        for q3 in grid_address[g3]:
            w.write("%4d %4d %4d    " % tuple(q3))
        w.write("\n")
    w.close()

def write_grid_address(grid_address, mesh, filename=None):
    grid_address_filename = "grid_address"
    suffix = "-m%d%d%d" % tuple(mesh)
    if filename is not None:
        suffix += "." + filename
    suffix += ".dat"
    grid_address_filename += suffix

    w = open(grid_address_filename, 'w')
    w.write("# Grid addresses for %dx%dx%d mesh\n" % tuple(mesh))
    w.write("#%9s    %8s %8s %8s     %8s %8s %8s\n" %
            ("index", "a", "b", "c",
             ("a%%%d" % mesh[0]), ("b%%%d" % mesh[1]), ("c%%%d" % mesh[2])))
    for i, bz_q in enumerate(grid_address):
        if i == np.prod(mesh):
            w.write("#" + "-" * 78 + "\n")
        q = bz_q % mesh
        w.write("%10d    %8d %8d %8d     " % (i, bz_q[0], bz_q[1], bz_q[2]))
        w.write("%8d %8d %8d\n" % tuple(q))

    return grid_address_filename

def write_grid_address_to_hdf5(grid_address,
                               mesh,
                               grid_mapping_table,
                               filename=None):
    suffix = _get_filename_suffix(mesh, filename=filename)
    full_filename = "grid_address" + suffix + ".hdf5"
    with h5py.File(full_filename, 'w') as w:
        w.create_dataset('mesh', data=mesh)
        w.create_dataset('grid_address', data=grid_address)
        w.create_dataset('grid_mapping_table', data=grid_mapping_table)
        return full_filename
    return None

def write_freq_shifts_to_hdf5(freq_shifts, filename='freq_shifts.hdf5'):
    with h5py.File(filename, 'w') as w:
        w.create_dataset('shift', data=freq_shifts)

def write_imag_self_energy_at_grid_point(gp,
                                         band_indices,
                                         mesh,
                                         frequencies,
                                         gammas,
                                         sigma=None,
                                         temperature=None,
                                         scattering_event_class=None,
                                         filename=None,
                                         is_mesh_symmetry=True):

    gammas_filename = "gammas"
    gammas_filename += "-m%d%d%d-g%d-" % (mesh[0],
                                          mesh[1],
                                          mesh[2],
                                          gp)
    if sigma is not None:
        gammas_filename += ("s%f" % sigma).rstrip('0').rstrip('\.') + "-"

    if temperature is not None:
        gammas_filename += ("t%f" % temperature).rstrip('0').rstrip('\.') + "-"

    for i in band_indices:
        gammas_filename += "b%d" % (i + 1)

    if scattering_event_class is not None:
        gammas_filename += "-c%d" % scattering_event_class

    if not filename == None:
        gammas_filename += ".%s" % filename
    elif not is_mesh_symmetry:
        gammas_filename += ".nosym"
    gammas_filename += ".dat"

    w = open(gammas_filename, 'w')
    for freq, g in zip(frequencies, gammas):
        w.write("%15.7f %20.15e\n" % (freq, g))
    w.close()

def write_joint_dos(gp,
                    mesh,
                    frequencies,
                    jdos,
                    sigma=None,
                    temperatures=None,
                    filename=None,
                    is_mesh_symmetry=True):
    if temperatures is None:
        _write_joint_dos_at_t(gp,
                              mesh,
                              frequencies,
                              jdos,
                              sigma=sigma,
                              temperature=None,
                              filename=filename,
                              is_mesh_symmetry=is_mesh_symmetry)
    else:
        for jdos_at_t, t in zip(jdos, temperatures):
            _write_joint_dos_at_t(gp,
                                  mesh,
                                  frequencies,
                                  jdos_at_t,
                                  sigma=sigma,
                                  temperature=t,
                                  filename=filename,
                                  is_mesh_symmetry=is_mesh_symmetry)

def _write_joint_dos_at_t(gp,
                          mesh,
                          frequencies,
                          jdos,
                          sigma=None,
                          temperature=None,
                          filename=None,
                          is_mesh_symmetry=True):
    jdos_filename = "jdos-m%d%d%d-g%d" % (mesh[0], mesh[1], mesh[2], gp)
    if sigma is not None:
        jdos_filename += ("-s%f" % sigma).rstrip('0').rstrip('\.')
    if temperature is not None:
        jdos_filename += ("-t%f" % temperature).rstrip('0').rstrip('\.')
    if not is_mesh_symmetry:
        jdos_filename += ".nosym"
    if filename is not None:
        jdos_filename += ".%s" % filename
    jdos_filename += ".dat"

    w = open(jdos_filename, 'w')
    for omega, vals in zip(frequencies, jdos):
        w.write("%15.7f" % omega)
        w.write((" %20.15e" * len(vals)) % tuple(vals))
        w.write("\n")
    w.close()

def write_linewidth_at_grid_point(gp,
                                  band_indices,
                                  temperatures,
                                  gamma,
                                  mesh,
                                  sigma=None,
                                  filename=None,
                                  is_mesh_symmetry=True):

    lw_filename = "linewidth"
    lw_filename += "-m%d%d%d-g%d-" % (mesh[0], mesh[1], mesh[2], gp)
    if sigma is not None:
        lw_filename += ("s%f" % sigma).rstrip('0') + "-"

    for i in band_indices:
        lw_filename += "b%d" % (i + 1)

    if not filename == None:
        lw_filename += ".%s" % filename
    elif not is_mesh_symmetry:
        lw_filename += ".nosym"
    lw_filename += ".dat"

    w = open(lw_filename, 'w')
    for v, t in zip(gamma.sum(axis=1) * 2 / gamma.shape[1], temperatures):
        w.write("%15.7f %20.15e\n" % (t, v))
    w.close()

def write_frequency_shift(gp,
                          band_indices,
                          temperatures,
                          delta,
                          mesh,
                          epsilon=None,
                          filename=None,
                          is_mesh_symmetry=True):

    fst_filename = "frequency_shift"
    fst_filename += "-m%d%d%d-g%d-" % (mesh[0], mesh[1], mesh[2], gp)
    if epsilon is not None:
        if epsilon > 1e-5:
            fst_filename += ("s%f" % epsilon).rstrip('0') + "-"
        else:
            fst_filename += ("s%.3e" % epsilon) + "-"
    for i in band_indices:
        fst_filename += "b%d" % (i + 1)
    if not filename == None:
        fst_filename += ".%s" % filename
    elif not is_mesh_symmetry:
        fst_filename += ".nosym"
    fst_filename += ".dat"

    w = open(fst_filename, 'w')
    for v, t in zip(delta.sum(axis=1) / delta.shape[1], temperatures):
        w.write("%15.7f %20.15e\n" % (t, v))
    w.close()

def write_collision_to_hdf5(temperature,
                            mesh,
                            gamma=None,
                            gamma_isotope=None,
                            collision_matrix=None,
                            grid_point=None,
                            band_index=None,
                            sigma=None,
                            sigma_cutoff=None,
                            filename=None):
    if band_index is None:
        band_indices = None
    else:
        band_indices = [band_index]
    suffix = _get_filename_suffix(mesh,
                                  grid_point=grid_point,
                                  band_indices=band_indices,
                                  sigma=sigma,
                                  sigma_cutoff=sigma_cutoff,
                                  filename=filename)
    full_filename = "collision" + suffix + ".hdf5"
    with h5py.File(full_filename, 'w') as w:
        w.create_dataset('temperature', data=temperature)
        if gamma is not None:
            w.create_dataset('gamma', data=gamma)
        if gamma_isotope is not None:
            w.create_dataset('gamma_isotope', data=gamma_isotope)
        if collision_matrix is not None:
            w.create_dataset('collision_matrix', data=collision_matrix)
        if grid_point is not None:
            w.create_dataset('grid_point', data=grid_point)
        if band_index is not None:
            w.create_dataset('band_index', data=(band_index + 1))
        if sigma is not None:
            w.create_dataset('sigma', data=sigma)
        if sigma_cutoff is not None:
            w.create_dataset('sigma_cutoff_width', data=sigma_cutoff)

        text = "Collisions "
        if grid_point is not None:
            text += "at grid adress %d " % grid_point
        if sigma is not None:
            if grid_point is not None:
                text += "and "
            else:
                text += "at "
            text += "sigma %s " % _del_zeros(sigma)
        text += "were written into "
        if sigma is not None:
            text += "\n"
        text += "\"%s\"." % ("collision" + suffix + ".hdf5")
        print(text)

    return full_filename

def write_full_collision_matrix(collision_matrix, filename='fcm.hdf5'):
    with h5py.File(filename, 'w') as w:
        w.create_dataset('collision_matrix', data=collision_matrix)

def write_unitary_matrix_to_hdf5(temperature,
                                 mesh,
                                 unitary_matrix=None,
                                 sigma=None,
                                 sigma_cutoff=None,
                                 filename=None):
    suffix = _get_filename_suffix(mesh,
                                  sigma=sigma,
                                  sigma_cutoff=sigma_cutoff,
                                  filename=filename)
    hdf5_filename = "unitary" + suffix + ".hdf5"
    with h5py.File(hdf5_filename, 'w') as w:
        w.create_dataset('temperature', data=temperature)
        if unitary_matrix is not None:
            w.create_dataset('unitary_matrix', data=unitary_matrix)

        if len(temperature) > 1:
            text = "Unitary matrices "
        else:
            text = "Unitary matrix "
        if sigma is not None:
            text += "at sigma %s" % _del_zeros(sigma)
            if sigma_cutoff is not None:
                text += "(%4.2f SD)" % sigma_cutoff
        if len(temperature) > 1:
            text += " were written into "
        else:
            text += " was written into "
        if sigma is not None:
            text += "\n"
        text += "\"%s\"." % hdf5_filename
        print(text)

def write_collision_eigenvalues_to_hdf5(temperatures,
                                        mesh,
                                        collision_eigenvalues,
                                        sigma=None,
                                        sigma_cutoff=None,
                                        filename=None,
                                        verbose=True):
    suffix = _get_filename_suffix(mesh,
                                  sigma=sigma,
                                  sigma_cutoff=sigma_cutoff,
                                  filename=filename)
    with h5py.File("coleigs" + suffix + ".hdf5", 'w') as w:
        w.create_dataset('temperature', data=temperatures)
        w.create_dataset('collision_eigenvalues', data=collision_eigenvalues)
        w.close()

        if verbose:
            text = "Eigenvalues of collision matrix "
            if sigma is not None:
                text += "with sigma %s\n" % sigma
            text += "were written into "
            text += "\"%s\"" % ("coleigs" + suffix + ".hdf5")
            print(text)

def write_kappa_to_hdf5(temperature,
                        mesh,
                        frequency=None,
                        group_velocity=None,
                        gv_by_gv=None,
                        mean_free_path=None,
                        heat_capacity=None,
                        kappa=None,
                        mode_kappa=None,
                        kappa_RTA=None, # RTA calculated in LBTE
                        mode_kappa_RTA=None, # RTA calculated in LBTE
                        gamma=None,
                        gamma_isotope=None,
                        gamma_N=None,
                        gamma_U=None,
                        averaged_pp_interaction=None,
                        qpoint=None,
                        weight=None,
                        mesh_divisors=None,
                        grid_point=None,
                        band_index=None,
                        sigma=None,
                        sigma_cutoff=None,
                        kappa_unit_conversion=None,
                        filename=None,
                        verbose=True):
    if band_index is None:
        band_indices = None
    else:
        band_indices = [band_index]
    suffix = _get_filename_suffix(mesh,
                                  mesh_divisors=mesh_divisors,
                                  grid_point=grid_point,
                                  band_indices=band_indices,
                                  sigma=sigma,
                                  sigma_cutoff=sigma_cutoff,
                                  filename=filename)
    full_filename = "kappa" + suffix + ".hdf5"
    with h5py.File(full_filename, 'w') as w:
        w.create_dataset('temperature', data=temperature)
        w.create_dataset('mesh', data=mesh)
        if frequency is not None:
            w.create_dataset('frequency', data=frequency)
        if group_velocity is not None:
            w.create_dataset('group_velocity', data=group_velocity)
        if gv_by_gv is not None:
            w.create_dataset('gv_by_gv', data=gv_by_gv)
        if mean_free_path is not None:
            w.create_dataset('mean_free_path', data=mean_free_path)
        if heat_capacity is not None:
            w.create_dataset('heat_capacity', data=heat_capacity)
        if kappa is not None:
            w.create_dataset('kappa', data=kappa)
        if mode_kappa is not None:
            w.create_dataset('mode_kappa', data=mode_kappa)
        if kappa_RTA is not None:
            w.create_dataset('kappa_RTA', data=kappa_RTA)
        if mode_kappa_RTA is not None:
            w.create_dataset('mode_kappa_RTA', data=mode_kappa_RTA)
        if gamma is not None:
            w.create_dataset('gamma', data=gamma)
        if gamma_isotope is not None:
            w.create_dataset('gamma_isotope', data=gamma_isotope)
        if gamma_N is not None:
            w.create_dataset('gamma_N', data=gamma_N)
        if gamma_U is not None:
            w.create_dataset('gamma_U', data=gamma_U)
        if averaged_pp_interaction is not None:
            w.create_dataset('ave_pp', data=averaged_pp_interaction)
        if qpoint is not None:
            w.create_dataset('qpoint', data=qpoint)
        if weight is not None:
            w.create_dataset('weight', data=weight)
        if grid_point is not None:
            w.create_dataset('grid_point', data=grid_point)
        if band_index is not None:
            w.create_dataset('band_index', data=(band_index + 1))
        if sigma is not None:
            w.create_dataset('sigma', data=sigma)
        if sigma_cutoff is not None:
            w.create_dataset('sigma_cutoff_width', data=sigma_cutoff)
        if kappa_unit_conversion is not None:
            w.create_dataset('kappa_unit_conversion',
                             data=kappa_unit_conversion)

        if verbose:
            text = ""
            if kappa is not None:
                text += "Thermal conductivity and related properties "
            else:
                text += "Thermal conductivity related properties "
            if grid_point is not None:
                text += "at gp-%d " % grid_point
                if band_index is not None:
                    text += "and band_index-%d\n" % (band_index + 1)
            if sigma is not None:
                if grid_point is not None:
                    text += "and "
                else:
                    text += "at "
                text += "sigma %s" % sigma
                if sigma_cutoff is None:
                    text += "\n"
                else:
                    text += "(%4.2f SD)\n" % sigma_cutoff
                text += "were written into "
            else:
                text += "were written into "
                if band_index is None:
                    text += "\n"
            text += "\"%s\"" % ("kappa" + suffix + ".hdf5")
            print(text)

        return full_filename

def read_gamma_from_hdf5(mesh,
                         mesh_divisors=None,
                         grid_point=None,
                         band_index=None,
                         sigma=None,
                         sigma_cutoff=None,
                         filename=None,
                         verbose=True):
    if band_index is None:
        band_indices = None
    else:
        band_indices = [band_index]
    suffix = _get_filename_suffix(mesh,
                                  mesh_divisors=mesh_divisors,
                                  grid_point=grid_point,
                                  band_indices=band_indices,
                                  sigma=sigma,
                                  sigma_cutoff=sigma_cutoff,
                                  filename=filename)
    if not os.path.exists("kappa" + suffix + ".hdf5"):
        if verbose:
            print("%s not found." % ("kappa" + suffix + ".hdf5"))
        return None

    read_data = {}

    with h5py.File("kappa" + suffix + ".hdf5", 'r') as f:
        read_data['gamma'] = f['gamma'][:]
        for key in ('gamma_isotope',
                    'ave_pp',
                    'gamma_N',
                    'gamma_U'):
            if key in f.keys():
                if len(f[key].shape) > 0:
                    read_data[key] = f[key][:]
                else:
                    read_data[key] = f[key][()]
        if verbose:
            print("Read data from %s." % ("kappa" + suffix + ".hdf5"))

    return read_data

def read_collision_from_hdf5(mesh,
                             indices=None,
                             grid_point=None,
                             band_index=None,
                             sigma=None,
                             sigma_cutoff=None,
                             filename=None,
                             verbose=True):
    if band_index is None:
        band_indices = None
    else:
        band_indices = [band_index]
    suffix = _get_filename_suffix(mesh,
                                  grid_point=grid_point,
                                  band_indices=band_indices,
                                  sigma=sigma,
                                  sigma_cutoff=sigma_cutoff,
                                  filename=filename)
    full_filename = "collision" + suffix + ".hdf5"
    if not os.path.exists(full_filename):
        if verbose:
            print("%s not found." % full_filename)
        return None

    with h5py.File(full_filename, 'r') as f:
        if indices == 'all':
            colmat_shape = (1,) + f['collision_matrix'].shape
            collision_matrix = np.zeros(colmat_shape, dtype='double', order='C')
            gamma = np.array(f['gamma'][:], dtype='double', order='C')
            collision_matrix[0] = f['collision_matrix'][:]
            temperatures = np.array(f['temperature'][:], dtype='double')
        else:
            colmat_shape = (1, len(indices)) + f['collision_matrix'].shape[1:]
            collision_matrix = np.zeros(colmat_shape, dtype='double')
            gamma = np.array(f['gamma'][indices], dtype='double', order='C')
            collision_matrix[0] = f['collision_matrix'][indices]
            temperatures = np.array(f['temperature'][indices], dtype='double')

        if verbose:
            text = "Collisions "
            if band_index is None:
                if grid_point is not None:
                    text += "at grid point %d " % grid_point
            else:
                if grid_point is not None:
                    text += ("at (grid point %d, band index %d) " %
                             (grid_point, band_index))
            if sigma is not None:
                if grid_point is not None:
                    text += "and "
                else:
                    text += "at "
                text += "sigma %s" % _del_zeros(sigma)
                if sigma_cutoff is not None:
                    text += "(%4.2f SD)" % sigma_cutoff
            if band_index is None and grid_point is not None:
                text += " were read from "
                text += "\n"
            else:
                text += "\n"
                text += "were read from "
            text += "%s." % full_filename
            print(text)

        return collision_matrix, gamma, temperatures

    return None

def write_amplitude_to_hdf5(amplitude,
                            mesh,
                            grid_point,
                            triplet=None,
                            weight=None,
                            frequency=None,
                            eigenvector=None):
    suffix = "-m%d%d%d" % tuple(mesh)
    suffix += ("-g%d" % grid_point)
    with h5py.File("amplitude" + suffix + ".hdf5", 'w') as w:
        w.create_dataset('amplitude', data=amplitude)
        if triplet is not None:
            w.create_dataset('triplet', data=triplet)
        if weight is not None:
            w.create_dataset('weight', data=weight)
        if frequency is not None:
            w.create_dataset('frequency', data=frequency)
        if eigenvector is not None:
            w.create_dataset('eigenvector', data=eigenvector)

def read_amplitude_from_hdf5(amplitudes_at_q,
                             mesh,
                             grid_point):
    suffix = "-m%d%d%d" % tuple(mesh)
    suffix += ("-g%d" % grid_point)
    with h5py.File("amplitude" + suffix + ".hdf5", 'r') as f:
        amplitudes_at_q[:] = f['amplitudes'][:]
        return amplitudes_at_q
    return None

def write_gamma_detail_to_hdf5(detailed_gamma,
                               temperature,
                               mesh,
                               grid_point,
                               sigma,
                               triplets,
                               weights,
                               frequency_points=None,
                               filename=None):
    suffix = _get_filename_suffix(mesh,
                                  grid_point=grid_point,
                                  sigma=sigma,
                                  filename=filename)
    full_filename = "gamma_detail" + suffix + ".hdf5"

    with h5py.File(full_filename, 'w') as w:
        w.create_dataset('gamma_detail', data=detailed_gamma)
        w.create_dataset('temperature', data=temperature)
        w.create_dataset('mesh', data=mesh)
        w.create_dataset('triplet', data=triplets)
        w.create_dataset('weight', data=weights)
        if frequency_points is not None:
            w.create_dataset('frequency_point', data=frequency_points)
        return full_filename

def write_phonon_to_hdf5(frequency,
                         eigenvector,
                         grid_address,
                         mesh,
                          filename=None):
    suffix = _get_filename_suffix(mesh, filename=filename)
    full_filename = "phonon" + suffix + ".hdf5"

    with h5py.File(full_filename, 'w') as w:
        w.create_dataset('mesh', data=mesh)
        w.create_dataset('grid_address', data=grid_address)
        w.create_dataset('frequency', data=frequency)
        w.create_dataset('eigenvector', data=eigenvector)
        return full_filename

def read_phonon_from_hdf5(mesh,
                          filename=None,
                          verbose=True):
    suffix = _get_filename_suffix(mesh, filename=filename)
    hdf5_filename = "phonon" + suffix + ".hdf5"
    if not os.path.exists(hdf5_filename):
        if verbose:
            print("%s not found." % hdf5_filename)
        return (None, None, None, None, hdf5_filename)

    with h5py.File(hdf5_filename, 'r') as f:
        frequencies = np.array(f['frequency'][:], dtype='double', order='C')
        itemsize = frequencies.itemsize
        eigenvectors = np.array(f['eigenvector'][:],
                                dtype=("c%d" % (itemsize * 2)), order='C')
        mesh_in_file = np.array(f['mesh'][:], dtype='intc')
        grid_address = np.array(f['grid_address'][:], dtype='intc', order='C')
        return (frequencies,
                eigenvectors,
                mesh_in_file,
                grid_address,
                hdf5_filename)

    return (None, None, None, None, hdf5_filename)

def write_ir_grid_points(mesh,
                         mesh_divs,
                         grid_points,
                         coarse_grid_weights,
                         grid_address,
                         primitive_lattice):
    w = open("ir_grid_points.yaml", 'w')
    w.write("mesh: [ %d, %d, %d ]\n" % tuple(mesh))
    if mesh_divs is not None:
        w.write("mesh_divisors: [ %d, %d, %d ]\n" % tuple(mesh_divs))
    w.write("reciprocal_lattice:\n")
    for vec, axis in zip(primitive_lattice.T, ('a*', 'b*', 'c*')):
        w.write("- [ %12.8f, %12.8f, %12.8f ] # %2s\n" % (tuple(vec) + (axis,)))
    w.write("num_reduced_ir_grid_points: %d\n" % len(grid_points))
    w.write("ir_grid_points:  # [address, weight]\n")

    for g, weight in zip(grid_points, coarse_grid_weights):
        w.write("- grid_point: %d\n" % g)
        w.write("  weight: %d\n" % weight)
        w.write("  grid_address: [ %12d, %12d, %12d ]\n" %
                tuple(grid_address[g]))
        w.write("  q-point:      [ %12.7f, %12.7f, %12.7f ]\n" %
                tuple(grid_address[g].astype('double') / mesh))

def parse_disp_fc2_yaml(filename="disp_fc2.yaml"):
    dataset = _parse_yaml(filename)
    natom = dataset['natom']
    new_dataset = {}
    new_dataset['natom'] = natom
    new_first_atoms = []
    for first_atoms in dataset['first_atoms']:
        first_atoms['number'] -= 1
        atom1 = first_atoms['number']
        disp1 = first_atoms['displacement']
        new_first_atoms.append({'number': atom1, 'displacement': disp1})
    new_dataset['first_atoms'] = new_first_atoms

    return new_dataset

def parse_disp_fc3_yaml(filename="disp_fc3.yaml"):
    dataset = _parse_yaml(filename)
    natom = dataset['natom']
    new_dataset = {}
    new_dataset['natom'] = natom
    if 'cutoff_distance' in dataset:
        new_dataset['cutoff_distance'] = dataset['cutoff_distance']
    new_first_atoms = []
    for first_atoms in dataset['first_atoms']:
        first_atoms['number'] -= 1
        atom1 = first_atoms['number']
        disp1 = first_atoms['displacement']
        new_second_atoms = []
        for second_atom in first_atoms['second_atoms']:
            second_atom['number'] -= 1
            atom2 = second_atom['number']
            if 'included' in second_atom:
                included = second_atom['included']
            else:
                included = True
            for disp2 in second_atom['displacements']:
                new_second_atoms.append({'number': atom2,
                                         'displacement': disp2,
                                         'included': included})
        new_first_atoms.append({'number': atom1,
                                'displacement': disp1,
                                'second_atoms': new_second_atoms})
    new_dataset['first_atoms'] = new_first_atoms

    return new_dataset

def parse_disp_fc4_yaml(filename="disp_fc4.yaml"):
    dataset = _parse_yaml(filename)
    natom = dataset['natom']
    new_dataset = {}
    new_dataset['natom'] = natom
    new_first_atoms = []
    for first_atoms in dataset['first_atoms']:
        first_atoms['number'] -= 1
        atom1 = first_atoms['number']
        disp1 = first_atoms['displacement']
        new_second_atoms = []
        for second_atoms in first_atoms['second_atoms']:
            second_atoms['number'] -= 1
            atom2 = second_atoms['number']
            disp2 = second_atoms['displacement']
            new_third_atoms = []
            for third_atoms in second_atoms['third_atoms']:
                third_atoms['number'] -= 1
                atom3 = third_atoms['number']
                for disp3 in third_atoms['displacements']:
                    new_third_atoms.append(
                        {'number': atom3, 'displacement': disp3})
            new_second_atoms.append(
                {'number': atom2,
                 'displacement': disp2,
                 'third_atoms': new_third_atoms})
        new_first_atoms.append(
            {'number': atom1,
             'displacement': disp1,
             'second_atoms': new_second_atoms})
    new_dataset['first_atoms'] = new_first_atoms

    new_dataset['num_first_displacements'] = dataset['num_first_displacements']
    new_dataset['num_second_displacements'] = dataset['num_second_displacements']
    new_dataset['num_third_displacements'] = dataset['num_third_displacements']

    return new_dataset

def parse_FORCES_FC2(disp_dataset, filename="FORCES_FC2"):
    num_atom = disp_dataset['natom']
    num_disp = len(disp_dataset['first_atoms'])
    forces_fc2 = []
    with open(filename, 'r') as f2:
        for i in range(num_disp):
            forces = _parse_force_lines(f2, num_atom)
            if forces is None:
                return []
            else:
                forces_fc2.append(forces)
    return forces_fc2

def parse_FORCES_FC3(disp_dataset, filename="FORCES_FC3"):
    num_atom = disp_dataset['natom']
    num_disp = len(disp_dataset['first_atoms'])
    for disp1 in disp_dataset['first_atoms']:
        num_disp += len(disp1['second_atoms'])

    forces_fc3 = []
    with open(filename, 'r') as f3:
        for i in range(num_disp):
            forces = _parse_force_lines(f3, num_atom)
            if forces is None:
                return []
            else:
                forces_fc3.append(forces)
    return forces_fc3

def parse_FORCES_FC4(disp_dataset, filename="FORCES_FC4"):
    num_atom = disp_dataset['natom']
    num_disp = len(disp_dataset['first_atoms'])
    for disp1 in disp_dataset['first_atoms']:
        num_disp += len(disp1['second_atoms'])
        for disp2 in disp1['second_atoms']:
            num_disp += len(disp2['third_atoms'])

    assert num_disp == (disp_dataset['num_first_displacements'] +
                        disp_dataset['num_second_displacements'] +
                        disp_dataset['num_third_displacements'])


    f4 = open(filename, 'r')
    forces_fc4 = [_parse_force_lines(f4, num_atom) for i in range(num_disp)]
    f4.close()
    return forces_fc4

def parse_QPOINTS3(filename='QPOINTS3'):
    f = open(filename)
    num = int(f.readline().strip())
    count = 0
    qpoints3 = []
    for line in f:
        line_array = [float(x) for x in line.strip().split()]

        if len(line_array) < 9:
            print("QPOINTS3 format is invalid.")
            raise ValueError
        else:
            qpoints3.append(line_array[0:9])

        count += 1
        if count == num:
            break

    return np.array(qpoints3)

def parse_fc3(num_atom, filename='fc3.dat'):
    f = open(filename)
    fc3 = np.zeros((num_atom, num_atom, num_atom, 3, 3, 3), dtype=float)
    for i in range(num_atom):
        for j in range(num_atom):
            for k in range(num_atom):
                f.readline()
                for l in range(3):
                    fc3[i, j, k, l] = [
                        [float(x) for x in f.readline().split()],
                        [float(x) for x in f.readline().split()],
                        [float(x) for x in f.readline().split()]]
                    f.readline()
    return fc3

def parse_fc2(num_atom, filename='fc2.dat'):
    f = open(filename)
    fc2 = np.zeros((num_atom, num_atom, 3, 3), dtype=float)
    for i in range(num_atom):
        for j in range(num_atom):
            f.readline()
            fc2[i, j] = [[float(x) for x in f.readline().split()],
                         [float(x) for x in f.readline().split()],
                         [float(x) for x in f.readline().split()]]
            f.readline()

    return fc2

def parse_triplets(filename):
    f = open(filename)
    triplets = []
    weights = []
    for line in f:
        if line.strip()[0] == "#":
            continue

        line_array = [int(x) for x in line.split()]
        triplets.append(line_array[:3])
        weights.append(line_array[3])

    return np.array(triplets), np.array(weights)

def parse_grid_address(filename):
    f = open(filename, 'r')
    grid_address = []
    for line in f:
        if line.strip()[0] == "#":
            continue

        line_array = [int(x) for x in line.split()]
        grid_address.append(line_array[1:4])

    return np.array(grid_address)

def _get_filename_suffix(mesh,
                         mesh_divisors=None,
                         grid_point=None,
                         band_indices=None,
                         sigma=None,
                         sigma_cutoff=None,
                         filename=None):
    suffix = "-m%d%d%d" % tuple(mesh)
    if mesh_divisors is not None:
        if (np.array(mesh_divisors, dtype=int) != 1).any():
            suffix += "-d%d%d%d" % tuple(mesh_divisors)
    if grid_point is not None:
        suffix += ("-g%d" % grid_point)
    if band_indices is not None:
        suffix += "-"
        for bi in band_indices:
            suffix += "b%d" % (bi + 1)
    if sigma is not None:
        suffix += "-s" + _del_zeros(sigma)
        if sigma_cutoff is not None:
            sigma_cutoff_str = _del_zeros(sigma_cutoff)
            suffix += "-sd" + sigma_cutoff_str
    if filename is not None:
        suffix += "." + filename

    return suffix

def _del_zeros(val):
    return ("%f" % val).rstrip('0').rstrip('\.')

def _parse_yaml(file_yaml):
    import yaml
    try:
        from yaml import CLoader as Loader
        from yaml import CDumper as Dumper
    except ImportError:
        from yaml import Loader, Dumper

    string = open(file_yaml).read()
    data = yaml.load(string, Loader=Loader)
    return data

def _parse_force_lines(forcefile, num_atom):
    forces = []
    for line in forcefile:
        if line.strip() == '':
            continue
        if line.strip()[0] == '#':
            continue
        forces.append([float(x) for x in line.strip().split()])
        if len(forces) == num_atom:
            break

    if not len(forces) == num_atom:
        return None
    else:
        return np.array(forces)

def _parse_force_constants_lines(fcthird_file, num_atom):
    fc2 = []
    for line in fcthird_file:
        if line.strip() == '':
            continue
        if line.strip()[0] == '#':
            continue
        fc2.append([float(x) for x in line.strip().split()])
        if len(fc2) == num_atom ** 2 * 3:
            break

    if not len(fc2) == num_atom ** 2 * 3:
        return None
    else:
        return np.array(fc2).reshape(num_atom, num_atom, 3, 3)
